# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Coverage check routine for RGBA center radius coverage checks."""

import logging

import numpy
from geoips.plugins.modules.coverage_checkers.center_radius import create_radius

LOG = logging.getLogger(__name__)

interface = "coverage_checkers"
family = "standard"
name = "center_radius_rgba"


def call(
    xarray_obj,
    variable_name,
    area_def=None,
    radius_km=300,
):
    """Coverage check routine for xarray objects with masked projected arrays.

    Only calculates coverage within a "radius_km" radius of center.

    Parameters
    ----------
    xarray_obj : xarray.Dataset
        xarray object containing variable "variable_name"
    variable_name : str
        variable name to check percent unmasked
        radius_km (float) : Radius of center disk to check for coverage

    Returns
    -------
    float
        Percent coverage of variable_name
    """
    if variable_name not in xarray_obj:
        raise KeyError(
            f"Variable {variable_name} did not exist. Can not calculate coverage."
        )

    temp_arr = xarray_obj[variable_name][:, :, 3]

    res_km = (
        min(
            xarray_obj.area_definition.pixel_size_x,
            xarray_obj.area_definition.pixel_size_y,
        )
        / 1000.0
    )
    radius_pixels = 1.0 * radius_km / res_km
    LOG.info(
        "Using %s km radius, %s pixels radius, %s km resolution, area_def %s",
        radius_km,
        radius_pixels,
        res_km,
        area_def,
    )

    dumby_arr = create_radius(
        temp_arr,
        radius_pixels=radius_pixels,
        x_center=temp_arr.shape[0] / 2,
        y_center=temp_arr.shape[1] / 2,
    )

    num_valid_in_radius = numpy.count_nonzero(
        numpy.logical_and(numpy.where(dumby_arr, 1, 0), numpy.where(temp_arr, 1, 0))
    )
    num_total_in_radius = numpy.count_nonzero(dumby_arr)
    return (float(num_valid_in_radius) / num_total_in_radius) * 100.0



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Test Order-based procflow product building classes."""

# Python Standard Libraries
# import copy

# Third-Party Libraries
# from pydantic import ValidationError
import pytest

# GeoIPS Libraries
from geoips.pydantic import workflows


# The following tests would be added in the later stage of OBP
# Dependent on further steps development


# def  test_product_spec_model_initialization():
"""Test to validate sequence of multiple steps definition."""

# def test_product_plugin_model_valid_product():
"""Test to validate a product definition."""


def test_bad_get_plugin_types_missing_types(valid_plugin_kinds):
    """Test get_plugin_kinds call to check there are no missing plugin kinds."""
    assert not (
        set(valid_plugin_kinds) - set(workflows.get_plugin_kinds())
    ), "Missing plugin kind(s)"


def test_good_get_plugin_names_valid_kind():
    """Test get_plugin_names call with a valid plugin kind."""
    plugin_names = workflows.get_plugin_names("reader")

    assert isinstance(plugin_names, list)
    for plugin in plugin_names:
        assert isinstance(plugin, str)
        # assert plugin


def test_bad_get_plugin_names_invalid_kind(caplog):
    """Test get_plugin_names call with an invalid plugin kind."""
    with caplog.at_level("CRITICAL"):
        with pytest.raises(AttributeError) as exec_info:
            workflows.get_plugin_names("invalid_plugin_kind")

        assert "is not a recognized plugin kind." in str(exec_info.value)

        # Check that critical log was made
        assert any(
            "is not a recognized plugin kind." in message for message in caplog.messages
        )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS formatted titles for tropical cyclone products."""

# Python Standard Libraries
import logging

LOG = logging.getLogger(__name__)

interface = "title_formatters"
family = "standard"
name = "tc_standard"


def call(
    area_def,
    xarray_obj,
    product_name_title,
    product_datatype_title=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    title_copyright=None,
):
    """Create standard GeoIPS formatted title for tropical cyclone products."""
    LOG.info("Setting dynamic title")

    # Make sure we reflect the actual start_datetime in the filename
    # geoimg_obj.set_geoimg_attrs(start_dt=xarray_obj.start_datetime)
    if "interpolated_time" in area_def.sector_info:
        sector_time = area_def.sector_info["interpolated_time"]
    else:
        sector_time = area_def.sector_info["synoptic_time"]

    title_line1 = "{0}{1:02d} {2} at {3}".format(
        area_def.sector_info["storm_basin"],
        int(area_def.sector_info["storm_num"]),
        area_def.sector_info["storm_name"],
        sector_time.strftime("%Y-%m-%d %H:%M:%S"),
    )

    # data_time = xarray_obj.start_datetime +
    #                            (xarray_obj.end_datetime - xarray_obj.start_datetime)/2
    data_time = xarray_obj.start_datetime
    # pandas dataframes seem to handle time objects much better than xarray.
    title_line2 = "{0} {1} at {2}".format(
        product_datatype_title,
        product_name_title,
        data_time.strftime("%Y-%m-%d %H:%M:%S"),
    )
    if bg_xarray is not None:
        # bg_data_time = bg_xarray.start_datetime +
        #                          (bg_xarray.end_datetime - bg_xarray.start_datetime)/2
        bg_data_time = bg_xarray.start_datetime
        title_line3 = "{0} {1} at {2}".format(
            bg_datatype_title,
            bg_product_name_title,
            bg_data_time.strftime("%Y-%m-%d %H:%M:%S"),
        )
        # title_string = f'{title_line1}\n{title_line2}\n{title_line3}\n
        #                                                             {title_copyright}'
        title_string = f"{title_line1}, {title_copyright}\n{title_line2}\n{title_line3}"
    else:
        # title_string = f'{title_line1}\n{title_line2}\n{title_copyright}'
        title_string = f"{title_line1}, {title_copyright}\n{title_line2}"
    LOG.info("title_string: %s", title_string)

    return title_string



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Default YAML metadata output format."""

import logging

from geoips.filenames.base_paths import PATHS as gpaths
from geoips.geoips_utils import replace_geoips_paths
from geoips.sector_utils.yaml_utils import write_yamldict

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "standard_metadata"
name = "metadata_default"


def call(
    area_def,
    xarray_obj,
    metadata_yaml_filename,
    product_filename,
    metadata_dir="metadata",
    basedir=gpaths["TCWWW"],
    output_dict=None,
    include_metadata_filename=False,
):
    """Produce metadata yaml file of sector info associated with final_product.

    Parameters
    ----------
    area_def : AreaDefinition
        Pyresample AreaDefintion object
    final_product : str
        Product that is associated with the passed area_def
    metadata_dir : str, default='metadata'
        Subdirectory name for metadata (using non-default allows for
        non-operational outputs)

    Returns
    -------
    str
        Metadata yaml filename, if one was produced.
    """
    from geoips.sector_utils.utils import is_sector_type

    if not is_sector_type(area_def, "tc"):
        return None
    # os.path.join does not take a list, so "*" it
    # product_partial_path = product_filename.replace(gpaths['TCWWW'],
    #   'https://www.nrlmry.navy.mil/tcdat')
    product_partial_path = replace_geoips_paths(product_filename)
    # product_partial_path = pathjoin(
    #   *final_product.split('/')[-5:-1]+[basename(final_product)])
    return output_metadata_yaml(
        metadata_yaml_filename,
        area_def,
        xarray_obj,
        product_partial_path,
        output_dict,
        include_metadata_filename=include_metadata_filename,
    )


def update_sector_info_with_default_metadata(
    area_def, xarray_obj, product_filename=None, metadata_filename=None
):
    """Update sector info found in "area_def" with standard metadata output.

    Parameters
    ----------
    area_def : AreaDefinition
        Pyresample AreaDefinition of sector information
    xarray_obj : xarray.Dataset
        xarray Dataset object that was used to produce product
    product_filename : str
        Full path to full product filename that this YAML file refers to

    Returns
    -------
    dict
        sector_info dict with standard metadata added
         * bounding box
         * product filename with wildcards
         * basename of original source filenames
    """
    sector_info = area_def.sector_info.copy()

    if hasattr(area_def, "sector_type") and "sector_type" not in sector_info:
        sector_info["sector_type"] = area_def.sector_type

    sector_info["bounding_box"] = {}
    sector_info["bounding_box"]["minlat"] = area_def.area_extent_ll[1]
    sector_info["bounding_box"]["maxlat"] = area_def.area_extent_ll[3]
    sector_info["bounding_box"]["minlon"] = area_def.area_extent_ll[0]
    sector_info["bounding_box"]["maxlon"] = area_def.area_extent_ll[2]
    sector_info["bounding_box"]["pixel_width_m"] = area_def.pixel_size_x
    sector_info["bounding_box"]["pixel_height_m"] = area_def.pixel_size_y
    sector_info["bounding_box"]["image_width"] = area_def.x_size
    sector_info["bounding_box"]["image_height"] = area_def.y_size
    sector_info["bounding_box"]["proj4_string"] = area_def.proj4_string

    if product_filename:
        sector_info["product_filename"] = replace_geoips_paths(product_filename)
    if metadata_filename:
        sector_info["metadata_filename"] = replace_geoips_paths(metadata_filename)

    if "source_file_names" in xarray_obj.attrs.keys():
        sector_info["source_file_names"] = xarray_obj.source_file_names
    # Backwards compatibility, so the default metadata doesn't change.
    if "source_file_names" in xarray_obj.attrs.keys():
        sector_info["source_file_names"] = xarray_obj.source_file_names

    return sector_info


def output_metadata_yaml(
    metadata_fname,
    area_def,
    xarray_obj,
    product_filename=None,
    output_dict=None,
    include_metadata_filename=False,
):
    """Write out yaml file "metadata_fname" of sector info found in "area_def".

    Parameters
    ----------
    metadata_fname : str
        Path to output metadata_fname
    area_def : AreaDefinition
        Pyresample AreaDefinition of sector information
    xarray_obj : xarray.Dataset
        xarray Dataset object that was used to produce product
    productname : str
        Full path to full product filename that this YAML file refers to

    Returns
    -------
    str
        Path to metadata filename if successfully produced.
    """
    sector_info_kwargs = {}
    if include_metadata_filename:
        sector_info_kwargs["metadata_filename"] = metadata_fname
    if product_filename:
        sector_info_kwargs["product_filename"] = product_filename
    sector_info = update_sector_info_with_default_metadata(
        area_def, xarray_obj, **sector_info_kwargs
    )

    returns = write_yamldict(
        sector_info, metadata_fname, force=True, replace_geoips_paths=True
    )
    if returns:
        LOG.info("METADATASUCCESS Writing %s", metadata_fname)
    return returns



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""TC product YAML metadata output format."""

import logging

from geoips.filenames.base_paths import PATHS as gpaths
from geoips.geoips_utils import replace_geoips_paths
from geoips.sector_utils.yaml_utils import write_yamldict
from geoips.interfaces import products

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "standard_metadata"
name = "metadata_tc"


def call(
    area_def,
    xarray_obj,
    metadata_yaml_filename,
    product_filename,
    metadata_dir="metadata",
    basedir=gpaths["TCWWW"],
    output_dict=None,
    metadata_fname_dict=None,
    output_fname_dict=None,
):
    """Produce metadata yaml file of sector info associated with final_product.

    Parameters
    ----------
    area_def : AreaDefinition
        Pyresample AreaDefintion object
    final_product : str
        Product that is associated with the passed area_def
    metadata_dir : str, default="metadata"
        Subdirectory name for metadata (using non-default allows for
        non-operational outputs)

    Returns
    -------
    str
        Metadata yaml filename, if one was produced.
    """
    from geoips.sector_utils.utils import is_sector_type

    if not is_sector_type(area_def, "tc"):
        return None
    # os.path.join does not take a list, so "*" it
    # product_partial_path = product_filename.replace(gpaths['TCWWW'],
    #   'https://www.nrlmry.navy.mil/tcdat')
    product_partial_path = replace_geoips_paths(product_filename)
    # product_partial_path = pathjoin(
    #   *final_product.split('/')[-5:-1]+[basename(final_product)])
    return output_tc_metadata_yaml(
        metadata_yaml_filename,
        area_def,
        xarray_obj,
        product_partial_path,
        output_dict,
        metadata_fname_dict=metadata_fname_dict,
        output_fname_dict=output_fname_dict,
    )


def update_sector_info_with_data_times(sector_info, xarray_obj):
    """Update sector info with data times, for YAML metadata output."""
    start = xarray_obj.start_datetime
    end = xarray_obj.end_datetime
    mid = start + (end - start) / 2.0
    sector_info["data_times"] = {}
    sector_info["data_times"]["start"] = start
    sector_info["data_times"]["end"] = end
    sector_info["data_times"]["mid"] = mid
    return sector_info


def update_sector_info_with_coverage(
    sector_info, product_name, xarray_obj, area_def, output_dict
):
    """Update sector info with coverage, for YAML metadata output."""
    from geoips.dev.product import get_covg_from_product, get_covg_args_from_product

    covg_func_types = {
        "image_production": "image_production_coverage_checker",
        "fname": "filename_coverage_checker",
        "full": "full_coverage_checker",
    }

    covg_funcs = {}
    covg_args = {}
    covgs = {}

    prod_plugin = products.get_plugin(
        xarray_obj.source_name, product_name, output_dict.get("product_spec_override")
    )

    default_covg_plugin = get_covg_from_product(
        prod_plugin,
        covg_field="coverage_checker",
    )

    default_covg_args = get_covg_args_from_product(
        prod_plugin,
        covg_field="coverage_checker",
    )
    try:
        default_covgs = default_covg_plugin(
            xarray_obj, prod_plugin.name, area_def, **covg_args["default"]
        )
    except KeyError:
        LOG.warning(
            '"%s" covg_func not defined, not including in metadata_tc output', "default"
        )

    for covg_func_type, coverage_checker_plugin_name in covg_func_types.items():
        covg_funcs[covg_func_type] = get_covg_from_product(
            prod_plugin,
            covg_field=coverage_checker_plugin_name,
        )
        covg_args[covg_func_type] = get_covg_args_from_product(
            prod_plugin,
            covg_field=coverage_checker_plugin_name,
        )
        # Set variable_name to prod_plugin.name if not defined.
        # Always use variable_name if it is defined.
        # Remove from args so there is not a duplicate are when passing
        # (since we are passing covg_varname explicitly).
        # Note get_covg_args_from_product was updated to return a copy of
        # covg_args, so this does not impact future uses of the product.
        covg_varname = covg_args[covg_func_type].pop("variable_name", prod_plugin.name)
        # Note variables can be specified as DATASET:VARIABLE, since this is a
        # preprocessed alg_xarray, and not a dictionary of datasets, just use the
        # variable name (we expect the correct variable will exist in this final
        # processed array)
        if ":" in covg_varname:
            covg_varname = covg_varname.split(":")[1]
        try:
            covgs[covg_func_type] = covg_funcs[covg_func_type](
                xarray_obj, covg_varname, area_def, **covg_args[covg_func_type]
            )
        except KeyError:
            LOG.warning(
                '"%s" covg_func not defined, not including in metadata_tc output',
                covg_func_type,
            )

    if covgs:
        sector_info["covg_info"] = {}

    for covg_func_type in covgs.keys():
        sector_info["covg_info"][covg_func_type + "_covg_func"] = covg_funcs[
            covg_func_type
        ].name
        sector_info["covg_info"][covg_func_type + "_covg_args"] = covg_args[
            covg_func_type
        ]
        sector_info["covg_info"][covg_func_type + "_covg"] = covgs[covg_func_type]

    if covgs.keys() and not set(covg_func_types).issubset(set(covgs.keys())):
        sector_info["covg_info"]["default_covg_func"] = default_covgs.name
        sector_info["covg_info"]["default_covg_args"] = default_covg_args
        sector_info["covg_info"]["default_covg"] = default_covgs

    return sector_info


def output_tc_metadata_yaml(
    metadata_fname,
    area_def,
    xarray_obj,
    product_filename=None,
    output_dict=None,
    metadata_fname_dict=None,
    output_fname_dict=None,
):
    """Write out yaml file "metadata_fname" of sector info found in "area_def".

    Parameters
    ----------
    metadata_fname : str
        Path to output metadata_fname
    area_def : AreaDefinition
        Pyresample AreaDefinition of sector information
    xarray_obj : xarray.Dataset
        xarray Dataset object that was used to produce product
    productname : str
        Full path to full product filename that this YAML file refers to

    Returns
    -------
    str
        Path to metadata filename if successfully produced.
    """
    from geoips.plugins.modules.output_formatters.metadata_default import (
        update_sector_info_with_default_metadata,
    )

    sector_info = update_sector_info_with_default_metadata(
        area_def, xarray_obj, product_filename=product_filename
    )

    sector_info = update_sector_info_with_coverage(
        sector_info,
        metadata_fname_dict["product_name"],
        xarray_obj,
        area_def,
        output_dict,
    )

    sector_info = update_sector_info_with_data_times(
        sector_info,
        xarray_obj,
    )

    returns = write_yamldict(
        sector_info, metadata_fname, force=True, replace_geoips_paths=True
    )
    if returns:
        LOG.info("METADATASUCCESS Writing %s", metadata_fname)
    return returns



"""Pydantic models used to validate GeoIPS feature annotator plugins."""

from pydantic import Field

from geoips.pydantic.bases import (
    FrozenModel,
    PermissiveFrozenModel,
    PluginModel,
    ColorType,
)


class CartopyFeature(PermissiveFrozenModel):
    """Generic model for cartopy features."""

    enabled: bool = Field(
        ..., strict=True, description="Whether or not to enable this feature."
    )
    edgecolor: ColorType = Field(
        None,
        description=(
            "A rgb tuple, matplotlib named color, or hexidecimal string (#XXXXXX) to "
            "apply to the edges of the cartopy feature.\n"
            "For more info, see: "
            "https://matplotlib.org/stable/users/explain/colors/colors.html"
        ),
    )
    # NOTE: Once we add land / ocean features, we'll need to add another field, labeled
    # facecolor.
    linewidth: float = Field(
        None, ge=0, description="The width in pixels of the specified feature."
    )


class FeatureAnnotatorSpec(FrozenModel):
    """Feature Annotator spec (specification) format."""

    # NOTE: Will need to add land and ocean features once they've been merged into main.
    coastline: CartopyFeature = Field(
        ..., strict=True, description="A cartopy coastline feature."
    )
    borders: CartopyFeature = Field(
        ..., strict=True, description="A cartopy borders feature."
    )
    rivers: CartopyFeature = Field(
        ..., strict=True, description="A cartopy rivers feature."
    )
    states: CartopyFeature = Field(
        ..., strict=True, description="A cartopy states feature."
    )
    background: ColorType = Field(
        None,
        description=(
            "A rgb tuple, matplotlib named color, or hexidecimal string (#XXXXXX) "
            "to apply to the background of your image.\n"
            "For more info, see: "
            "https://matplotlib.org/stable/users/explain/colors/colors.html"
        ),
    )


class FeatureAnnotatorPluginModel(PluginModel):
    """Feature Annotator plugin format."""

    spec: FeatureAnnotatorSpec = Field(
        ...,
        description=(
            "Specification of how to apply cartopy features to your annotated imagery. "
            "Works alongside matplotlib and cartopy to generate these features."
            "For more information, see: "
            "https://scitools.org.uk/cartopy/docs/v0.14/matplotlib/feature_interface.html"  # NOQAx
        ),
    )



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `validate` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

from glob import glob
from importlib import resources
import pytest

from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


class TestGeoipsValidate(BaseCliTest):
    """Unit Testing Class for 'geoips validate' Command."""

    @property
    def command_combinations(self):
        """A list of stochastic call signatures for the GeoipsValidate command.

        This includes failing cases as well.
        """
        if not hasattr(self, "_cmd_list"):
            self._cmd_list = []
            base_args = ["geoips", "validate"]
            # validate some subset plugins from all installed packages
            for pkg_name in self.plugin_package_names:
                pkg_path = str(resources.files(pkg_name) / "plugins")
                for plugin_type in ["modules", "yaml"]:
                    if plugin_type == "modules":
                        plugin_path_str = f"{pkg_path}/{plugin_type}/**/*.py"
                    else:
                        plugin_path_str = f"{pkg_path}/{plugin_type}/**/*.yaml"
                    plugin_paths = sorted(glob(plugin_path_str, recursive=True))
                    if plugin_type == "modules":
                        # Filter out '__init__.py' files and any file in a '/utils/' dir
                        plugin_paths = [
                            f
                            for f in plugin_paths
                            if not f.endswith("__init__.py")
                            and "/utils/" not in f.replace("\\", "/")
                        ]
                    # Just use every fifth plugin so we don't have random tests
                    for pidx in range(0, len(plugin_paths), 5):
                        # Not testing the validation of workflow plugins for the time
                        # being. Since these can be multi-document plugins, we have no
                        # way at runtime in unit testing to provide the name of the
                        # plugin in the multi-document plugin without hardcoding or
                        # searching through the plugin registry to find plugins that
                        # match the plugin path. Validation of these plugins is
                        # supported, but required the call to look like this instead
                        # 'geoips validate <workflow_path> <plugin_name>' if the file
                        # is indeed a multi-document.
                        plugin_path = plugin_paths[pidx]
                        if "workflows" not in plugin_path:
                            self._cmd_list.append(base_args + [plugin_path])
            # Add argument list to retrieve help message
            self._cmd_list.append(base_args + ["-h"])
        return self._cmd_list

    def check_error(self, args, error):
        """Ensure that the 'geoips validate ...' error output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI (expected to fail)
        error: str
            - Multiline str representing the error output of the CLI call
        """
        # An error occurred using args. Assert that args is not valid and check the
        # output of the error.
        assert "usage: To use, type `geoips validate <file_path>`" in error
        assert "is invalid." in error

    def check_output(self, args, output):
        """Ensure that the 'geoips validate ...' successful output is correct.

        Parameters
        ----------
        args: 2D list of str
            - The arguments used to call the CLI
        output: str
            - Multiline str representing the output of the CLI call
        """
        # The args provided are valid, so test that the output is actually correct
        if "-h" in args:
            assert "usage: To use, type `geoips validate <file_path>" in output
        else:
            # Checking that output from geoips validate command reports valid
            assert "is valid." in output


test_sub_cmd = TestGeoipsValidate()


@pytest.mark.parametrize(
    "args",
    test_sub_cmd.command_combinations,
    ids=test_sub_cmd.generate_id,
)
def test_command_combinations(monkeypatch, args):
    """Test all 'geoips validate ...' commands.

    This test covers every valid combination of commands for the 'geoips validate'
    command. We also test invalid commands, to ensure that the proper help documentation
    is provided for those using the command incorrectly.

    Parameters
    ----------
    args: 2D array of str
        - List of arguments to call the CLI with (ie. ['geoips', 'validate'])
    """
    test_sub_cmd.test_command_combinations(monkeypatch, args)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS CLI "list" command.

Lists the appropriate interfaces/packages/plugins based on the arguments provided.
"""

from glob import glob
from importlib import metadata, resources, import_module
import json
from os import listdir
from os.path import basename, exists
import sys

from tabulate import tabulate

from geoips.commandline.ancillary_info.test_data import test_dataset_dict
from geoips.commandline.geoips_command import (
    CommandClassFactory,
    GeoipsCommand,
    GeoipsExecutableCommand,
)
from geoips.geoips_utils import is_editable
from geoips import interfaces


class GeoipsListRegistries(GeoipsExecutableCommand):
    """List Command for listing out plugin registries from installed plugin packages.

    This command is used to easily expose what plugin packages have plugin registries
    created.
    """

    name = "registries"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-subparser for the List Registries Command."""
        self.parser.add_argument(
            "-n",
            "--namespace",
            type=str,
            default="geoips.plugin_packages",
            help=(
                "The namespace of plugin packages to list plugin registries for. "
                "If not specified, this defaults to 'geoips.plugin_packages'."
            ),
        )
        self.parser.add_argument(
            "-r",
            "--relpath",
            default=False,
            action="store_true",
            help=(
                "Whether or not you want the paths for the plugin registry files to be "
                "displayed as relative paths. Default is false, which means the paths "
                "will be shown as absolute."
            ),
        )

    def __call__(self, args):
        """List all plugin registries under 'namespace' held under <package_name>.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - GeoIPS Package
            - .json registry path
            - .yaml registry path

        Parameters
        ----------
        args: Namespace()
            - The list argument namespace to parse through
        """
        namespace = args.namespace
        relpath = args.relpath
        package_name = args.package_name
        default_headers = {
            "package": "GeoIPS Package",
            "json": "JSON Path",
            "yaml": "YAML Path",
        }
        headers = self._get_headers_by_command(args, default_headers)
        registry_list = []
        packages = metadata.entry_points(group=namespace)

        if len(packages) == 0:
            self.parser.error(
                f"Error: no plugin packages exist under namespace '{namespace}'."
            )

        for pkg in metadata.entry_points(group=namespace):
            pkg_path = resources.files(pkg.value)
            json_path = str(pkg_path / "registered_plugins.json")
            yaml_path = str(pkg_path / "registered_plugins.yaml")

            json_path = json_path if exists(json_path) else "missing"
            yaml_path = yaml_path if exists(yaml_path) else "missing"

            json_path = basename(json_path) if relpath else json_path
            yaml_path = basename(yaml_path) if relpath else yaml_path

            reg_entry = [pkg.value, json_path, yaml_path]

            if package_name != "all" and pkg.value == package_name:
                registry_list.append(reg_entry)
                break
            elif package_name == "all":
                registry_list.append(reg_entry)
            else:
                continue

        if len(registry_list) == 0:
            self.parser.error(
                f"Error: plugin package '{package_name}' could not be found in "
                f"namespace '{namespace}'."
            )

        print("-" * len("Plugin Package Registries"))
        print("Plugin Package Registries")
        print("-" * len("Plugin Package Registries"))
        print(
            tabulate(
                sorted(registry_list),
                headers=headers.values(),
                tablefmt="rounded_grid",
                maxcolwidths=self.terminal_width // len(headers),
            )
        )


class GeoipsListSourceNames(GeoipsExecutableCommand):
    """List Command for listing out source_names from available reader plugins.

    This command is used to easily expose what source_names are supported by each
    reader. If a reader module is missing the source_name attribute, then that will be
    notified in the CLI output.
    """

    name = "source-names"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-subparser for the List Source Names Command."""
        pass

    def __call__(self, args):
        """List all source-names in reader plugins held under <package_name>.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - GeoIPS Package
            - Source Names
            - Reader Name

        Parameters
        ----------
        args: Namespace()
            - The list argument namespace to parse through
        """
        package_name = args.package_name
        default_headers = {
            "source_name": "Source Name",
            "reader_names": "Reader Names",
        }
        interface_registry = self._get_registry_by_interface_and_package(
            interfaces.readers, package_name
        )
        if interface_registry is None:
            print(
                f"\nPlugin Package '{package_name}' has no reader plugins and "
                "therefore no implemented source_names."
            )
            return
        headers = self._get_headers_by_command(args, default_headers)
        src_name_mapping = self.match_readers_with_source_name(interface_registry)
        src_name_info = []
        for src_name, rdrs in src_name_mapping.items():
            src_name_entry = []
            for header in list(headers.keys()):
                if header == "source_name":
                    src_name_entry.append(src_name)
                elif header == "reader_names":
                    src_name_entry.append(list(rdrs))
            src_name_info.append(src_name_entry)
        if package_name == "all":
            package_name = "All GeoIPS Plugin Packages"
        print("-" * len(f"'{package_name}' Reader Source Names"))
        print(f"'{package_name}' Reader Source Names")
        print("-" * len(f"'{package_name}' Reader Source Names"))
        print(
            tabulate(
                sorted(src_name_info),
                headers=headers.values(),
                tablefmt="rounded_grid",
                maxcolwidths=self.terminal_width // len(headers),
            )
        )

    def match_readers_with_source_name(self, registry):
        """Assemble a dictionary of <source_name>:[readers] for tabular listing.

        Match source_names found in the registry to readers which support that
        source_name. Each key in the dictionary will be a unique source_name with values
        that are the names of readers which support that source_name.

        Parameters
        ----------
        registry: dict
            - The readers portion of all GeoIPS plugin packages' plugin registries or
              a single package's plugin registry.

        Returns
        -------
        src_name_mapping: dict
            - A dictionary mapping of source_name:[readers] which links readers to a
              unique source_name which they support.
        """
        src_name_mapping = {}
        for rdr_name in registry:
            src_names = registry[rdr_name]["source_names"]
            for src_name in src_names:
                if src_name not in src_name_mapping:
                    src_name_mapping[src_name] = set([rdr_name])
                else:
                    src_name_mapping[src_name].add(rdr_name)
        return src_name_mapping


class GeoipsListUnitTests(GeoipsExecutableCommand):
    """List Command for listing out available unit tests.

    Called via `geoips list unit-tests`. Outputs the following in a tabular format.
    """

    name = "unit-tests"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-subparser for the List Unit Tests Command."""
        pass

    def __call__(self, args):
        """List all of the available unit-tests held under <package_name>.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - Data Host
            - Dataset Name

        Parameters
        ----------
        args: Namespace()
            - The list argument namespace to parse through
        """
        package_name = args.package_name
        if package_name == "all":
            package_names = self.plugin_package_names
        else:
            package_names = [package_name]
        default_headers = {
            "package": "GeoIPS Package",
            "unit_test_dir": "Unit Test Directory",
            "unit_test_name": "Unit Test Name",
        }
        headers = self._get_headers_by_command(args, default_headers)
        for pkg_name in package_names:
            unit_test_info = []
            if not is_editable(pkg_name):
                # Package is installed in non-editable mode and we will not be able to
                # access unit tests. Raise a runtime error reporting this.
                print(
                    f"Error: Package '{pkg_name}' is installed in non-editable mode and"
                    " we are not able to access it's unit tests. For this command to "
                    f"work, please install '{pkg_name}' in editable mode via: "
                    f"'pip install -e <path_to_{pkg_name}>'",
                    file=sys.stderr,
                )
                # We use a print to sys.stderr so monkeypatch unit tests can catch this
                # output
                raise RuntimeError(
                    f"Package '{pkg_name}' isn't installed in editable mode."
                )
            unit_test_dir = str(resources.files(pkg_name) / "../tests/unit_tests")
            try:
                listdir(unit_test_dir)
            except FileNotFoundError:
                # I'm not sure why len(package_names) == 1 would indicate that the
                # unit-tests folder does not exist.  Instead, check if the unit-tests
                # folder exists to see if it exists.
                # if len(package_names) == 1:
                if not exists(unit_test_dir):
                    err_str = f"No unit test directory found under {pkg_name}. "
                    err_str += "Please create a tests/unit_tests folder for that "
                    err_str += "package if you want to continue."
                    self.parser.error(err_str)
                else:
                    print(f"package_names: {package_names}")
                    print(f"No unit tests found in '{pkg_name}', continuing.")
                    continue
            for subdir_name in listdir(unit_test_dir):
                for unit_test in sorted(
                    glob(f"{unit_test_dir}/{subdir_name}/test*.py")
                ):  # noqa
                    unit_test_entry = []
                    for header in list(headers.keys()):
                        if header == "package":
                            unit_test_entry.append(pkg_name)
                        elif header == "unit_test_dir":
                            unit_test_entry.append(subdir_name)
                        elif header == "unit_test_name":
                            unit_test_entry.append(basename(unit_test))
                    unit_test_info.append(unit_test_entry)
            print("-" * len(f"'{pkg_name}' Unit Tests"))
            print(f"'{pkg_name}' Unit Tests")
            print("-" * len(f"'{pkg_name}' Unit Tests"))
            print(
                tabulate(
                    unit_test_info,
                    headers=headers.values(),
                    tablefmt="rounded_grid",
                    # maxcolwidths=self.terminal_width // len(headers),
                )
            )


class GeoipsListTestDatasets(GeoipsExecutableCommand):
    """List Command for listing off available test-datasets used by GeoIPS.

    Called via `geoips list test-datasets`. Outputs the following in a tabular format.
    """

    name = "test-datasets"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-subparser for the List Test Datasets Command.

        Since this command requires no additional arguments, we pass for the time being.
        """
        pass

    def __call__(self, args):
        """List all of the test datasets used by GeoIPS.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - Data Host
            - Dataset Name

        Parameters
        ----------
        args: Namespace()
            - The list argument namespace to parse through
        """
        if args.package_name != "all":
            self.parser.error("Error: '-p' flag is not supported for this command")
        dataset_info = []
        default_headers = {"data_host": "Data Host", "dataset_name": "Dataset Name"}
        headers = self._get_headers_by_command(args, default_headers)
        for test_dataset_name in list(test_dataset_dict.keys()):
            dataset_entry = []
            for header in list(headers.keys()):
                if header == "data_host":
                    dataset_entry.append("io.cira.colostate.edu")
                elif header == "dataset_name":
                    dataset_entry.append(test_dataset_name)
            dataset_info.append(dataset_entry)
        print("-" * len("Available Test Datasets"))
        print("Available Test Datasets")
        print("-" * len("Available Test Datasets"))
        print(
            tabulate(
                dataset_info,
                headers=headers.values(),
                tablefmt="rounded_grid",
                maxcolwidths=self.terminal_width // len(headers),
            )
        )


class GeoipsListInterfaces(GeoipsExecutableCommand):
    """List Command for listing interfaces and interface-specific plugins.

    Will either list information about available GeoIPS interfaces (done via
    `geoips list interfaces`), or list information about implemented interface-specific
    plugins implemented in all, or a certain package (done via
    'geoips list interfaces -i <-p> <package_name>').

    Called via `geoips list interfaces`. Outputs the following data in a tabular format.
    """

    name = "interfaces"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-suparser for the List Interfaces Command."""
        self.parser.add_argument(
            "--implemented",
            "-i",
            default=False,
            action="store_true",
            help=str(
                "Flag to list what's implemented in each package, "
                + "rather than what's available."
            ),
        )

    def __call__(self, args):
        """List the available interface[s] within [a] GeoIPS Package[s]".

        Printed to Terminal (either available or implemented)
        -----------------------------------------------------
        available_out_array: 2D Array of Strings
            - Absolute Path
            - Docstring
            - GeoIPS Package
            - Interface Type
            - Interface Name
            - Supported Families
        implemented_out_array: 2D Array of Strings
            - GeoIPS Package
            - Interface Type
            - Interface Name

        Parameters
        ----------
        args: Argparse Namespace()
            - The Argument Namespace for GeoipsListInterfaces Command
        """
        package_name = args.package_name
        # Flag representing whether or not we want to list what's implemented or
        # what's available.
        implemented = args.implemented
        if package_name != "all" or (implemented and package_name):
            # If you're listing a certain package, implemented is implied. If it's
            # 'all' packages, make sure implemented as added if we are going to list in
            # that fasion.
            self.list_implemented_interfaces(package_name, args)
        else:
            # Otherwise just list off available interfaces.
            self.list_available_interfaces(args)

    def list_available_interfaces(self, args):
        """List the available interface[s] within [a] GeoIPS Package[s]".

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - Absolute Path
            - Docstring
            - GeoIPS Package
            - Interface Type
            - Interface Name
            - Supported Families

        Parameters
        ----------
        args: argparse Argument Namespace
            - The arguments provided to a certain list command
        """
        interface_data = []
        default_headers = {
            "package": "GeoIPS Package",
            "plugin_type": "Interface Type",
            "interface": "Interface Name",
            "supported_families": "Supported Families",
            "docstring": "Docstring",
            "abspath": "Absolute Path",
        }
        headers = self._get_headers_by_command(args, default_headers)
        for interface_name in interfaces.__all__:
            interface = getattr(interfaces, interface_name)
            interface_entry = []
            for header in list(headers.keys()):
                if header == "package":
                    interface_entry.append("geoips")
                elif header == "plugin_type":
                    interface_entry.append(interface.interface_type)
                elif header == "interface":
                    interface_entry.append(interface_name)
                elif header == "supported_families":
                    interface_entry.append(",\n".join(interface.supported_families))
                elif header == "docstring":
                    interface_entry.append(interface.__doc__.split("\n")[0])
                elif header == "abspath":
                    interface_entry.append(
                        str(
                            resources.files("geoips")
                            / f"interfaces/{interface.interface_type}/{interface.name}.py"  # noqa
                        ),
                    )
            interface_data.append(interface_entry)

        print("-" * len("GeoIPS Interfaces"))
        print("GeoIPS Interfaces")
        print("-" * len("GeoIPS Interfaces"))
        print(
            tabulate(
                interface_data,
                headers=headers.values(),
                tablefmt="rounded_grid",
                maxcolwidths=self.terminal_width // len(headers),
            )
        )

    def list_implemented_interfaces(self, package_name, args):
        """List the implemented interface[s] within [a] GeoIPS Package[s].

        Ie. search through all, or an individual package and list off what has been
        implemented in such package[s].

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - GeoIPS Package
            - Interface Type
            - Interface Name

        Parameters
        ----------
        package_name: str
            - The GeoIPS Package name whose scripts you want to list.
        args: argparse Argument Namespace
            - The arguments provided to a certain list command
        """
        default_headers = {
            "package": "GeoIPS Package",
            "plugin_type": "Interface Type",
            "interface": "Interface Name",
        }
        headers = self._get_headers_by_command(args, default_headers)
        for plugin_package_name, pkg_path in zip(
            self.plugin_package_names, self.plugin_package_paths
        ):

            if package_name == "all" or package_name == plugin_package_name:
                with open(
                    f"{pkg_path}/{plugin_package_name}/registered_plugins.json", "r"
                ) as fo:
                    pkg_registry = json.load(fo)
            else:
                continue
            interface_data = []
            for interface_type in pkg_registry.keys():
                for interface_name in pkg_registry[interface_type].keys():
                    interface_entry = []
                    for header in list(headers.keys()):
                        if header == "package":
                            interface_entry.append(plugin_package_name)
                        elif header == "plugin_type":
                            interface_entry.append(interface_type)
                        elif header == "interface":
                            interface_entry.append(interface_name)
                    interface_data.append(interface_entry)
            print("-" * len(f"{plugin_package_name.title()} Interfaces"))
            print(f"{plugin_package_name.title()} Interfaces")
            print("-" * len(f"{plugin_package_name.title()} Interfaces"))
            if len(interface_data) == 0:
                print(f"Package '{plugin_package_name.title()}' has no interfaces.")
            else:
                print(
                    tabulate(
                        interface_data,
                        headers=headers.values(),
                        tablefmt="rounded_grid",
                        maxcolwidths=self.terminal_width // len(headers),
                    )
                )


class GeoipsListPackages(GeoipsExecutableCommand):
    """List Command for listing off installed GeoIPS Packages.

    Called via `geoips list packages`. Outputs the following data in a tabular format.
    """

    name = "packages"
    command_classes = []

    def add_arguments(self):
        """No arguments for the list-subparser for the List Packages Command."""
        pass

    def __call__(self, args):
        """List all of the available GeoIPS Packages.

        Only installed packages will be listed. Ie, if geoips_clavrx, or another
        suitable package exists, but is not installed, it will not be listed.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - Docstring
            - GeoIPS Package
            - Package Path
            - Version Number

        Parameters
        ----------
        args: Argparse Namespace()
            - The Argument Namespace for GeoipsListPackages Command
        """
        if args.package_name != "all":
            # 'all' is the default for package name. We don't support the -p argument
            # for this command so we need to raise an error
            self.parser.error("Error: '-p' flag is not supported for this command")
        pkg_data = []
        default_headers = {
            "package": "GeoIPS Package",
            "docstring": "Docstring",
            "package_path": "Package Path",
            "version": "Version Number",
        }
        headers = self._get_headers_by_command(args, default_headers)
        pkg_name_requested = False
        for package_name, package_path in zip(
            self.plugin_package_names, self.plugin_package_paths
        ):

            pkg_entry = []
            docstring = import_module(package_name).__doc__
            for header in headers:
                if header == "package":
                    pkg_name_requested = True
                    pkg_entry.append(package_name)
                elif header == "docstring":
                    pkg_entry.append(str(docstring))
                elif header == "package_path":
                    pkg_entry.append(package_path)
                elif header == "version":
                    pkg_entry.append(metadata.version(package_name))
            pkg_data.append(pkg_entry)

        if pkg_name_requested:
            # If package name is in the list of headers, make sure we don't wrap that
            # name. This will cause failed tests and this is a workaround that ensures
            # the package name is not broken into multiple lines / words.
            colwidths = [None] + (len(headers) - 1) * [
                self.terminal_width // (len(headers))
            ]
        else:
            # Otherwise just keep it as normal.
            colwidths = self.terminal_width // len(headers)

        print("-" * len("GeoIPS Packages"))
        print("GeoIPS Packages")
        print("-" * len("GeoIPS Packages"))
        print(
            tabulate(
                pkg_data,
                headers=headers.values(),
                tablefmt="rounded_grid",
                maxcolwidths=colwidths,
            )
        )


class GeoipsListPlugins(GeoipsExecutableCommand):
    """List Command for listing off plugins in all, or a certain GeoIPS Package.

    Called via `geoips list plugins`. Outputs the following data in a tabular format.
    """

    name = "plugins"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-subparser for the List Plugins Command."""
        pass

    def __call__(self, args):
        """List the available interface[s] and their corresponding plugin names.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - Family Name
            - GeoIPS Package
            - Interface Name
            - Interface Type
            - Plugin Name
            - Relative Path

        Parameters
        ----------
        args: Argparse Namespace()
            - The list argument namespace to parse through
        """
        package_name = args.package_name
        # List of every available interface
        interfaces_to_list = [
            getattr(interfaces, name) for name in sorted(interfaces.__all__)
        ]

        plugin_found = False
        for curr_interface in interfaces_to_list:
            interface_registry = self._get_registry_by_interface_and_package(
                curr_interface, package_name
            )
            # Interface Registry can be None if we are looking through a specific
            # package whose registry doesn't contain that certain interface.
            if interface_registry is None:
                continue
            else:
                plugin_found = True
            print("-" * len(curr_interface.name))
            print(curr_interface.name)
            print("-" * len(curr_interface.name))
            self._print_plugins(curr_interface, interface_registry, args)
        if not plugin_found:
            print(f"Plugin package '{package_name.title()}' has no plugins.")


class GeoipsListSingleInterface(GeoipsExecutableCommand):
    """List Command for listing plugins of a single interface."""

    name = "interface"
    command_classes = []

    def add_arguments(self):
        """Instantiate the valid arguments that are supported for the list command.

        Currently the "geoips list interface" command supports this format:
            - geoips list interface <interface_name> -p <package_name>
        Where:
            - <interface_name> is any of the GeoIPS Interfaces' Name
            - <package_name> is any GeoIPS package that is installed and recognized by
              the GeoIPS Library
        """
        pass

    def __call__(self, args):
        """List all elements of the selected list option.

        Selected list option is args.interface_name, where interface_name can be any of:
            - any geoips.interface

        Where any of those options can be GeoIPS Package specific or from any package.

        Printed to Terminal
        -------------------
        out_array: 2D array of Strings
            - Family Name
            - GeoIPS Package
            - Interface Name
            - Interface Type
            - Plugin Name
            - Relative Path

        Parameters
        ----------
        args: Namespace()
            - The list argument namespace to parse through
        """
        interface_name = self.name.replace("-", "_")
        package_name = args.package_name
        try:
            interface = getattr(interfaces, interface_name)
        except AttributeError:
            self.parser.error(
                f"Interface: {interface_name} doesn't exist. Provide a valid interface."
            )
        interface_registry = self._get_registry_by_interface_and_package(
            interface, package_name
        )
        # Interface Registry can be None if we are looking through a specific
        # package whose registry doesn't contain that certain interface.
        if interface_registry is None:
            info_str = f"No plugins found under interface `{interface_name}` under "
            info_str += f"package `{package_name}`."
            print(info_str)
        else:
            print("-" * len(interface_name))
            print(interface_name)
            print("-" * len(interface_name))
            self._print_plugins(interface, interface_registry, args)


class GeoipsListScripts(GeoipsExecutableCommand):
    """List Command for listing test scripts from all, or a certain GeoIPS Package.

    Called via `geoips list scripts`. Outputs the following data in a tabular format.
    """

    name = "scripts"
    command_classes = []

    def add_arguments(self):
        """Add arguments to the list-subparser for the List Scripts Command."""
        pass

    def __call__(self, args):
        """List all of the available scripts held under <package_name>.

        Printed to Terminal
        -------------------
        out_array: 2D Array of Strings
            - GeoIPS Package
            - Script Name

        Parameters
        ----------
        args: Namespace()
            - The list argument namespace to parse through
        """
        package_name = args.package_name
        # columns = args.columns
        if package_name == "all":
            # list scripts found throughout all packages.
            plugin_package_names = self.plugin_package_names
        else:
            # list scripts from a certain package.
            plugin_package_names = [package_name]
        default_headers = {"package": "GeoIPS Package", "filename": "Filename"}
        headers = self._get_headers_by_command(args, default_headers)
        for plugin_package_name in plugin_package_names:
            if not is_editable(plugin_package_name):
                # Package is installed in non-editable mode and we will not be able to
                # access unit tests. Raise a runtime error reporting this.
                print(
                    f"Error: Package '{plugin_package_name}' is installed in "
                    "non-editable mode and we are not able to access it's unit tests.\n"
                    f"For this command to work, please install '{plugin_package_name}' "
                    "in editable mode via:\n'pip install -e "
                    f"<path_to_{plugin_package_name}>'",
                    file=sys.stderr,
                )
                # We use a print to sys.stderr so monkeypatch unit tests can catch this
                # output
                raise RuntimeError(
                    f"Package '{plugin_package_name}' isn't installed in editable mode."
                )
            script_names = sorted(
                [
                    [plugin_package_name, basename(fpath)]
                    for fpath in glob(
                        str(
                            resources.files(plugin_package_name)
                            / "../tests/scripts"
                            / "*.sh"
                        )
                    )
                ]
            )
            if headers != default_headers:
                for script_idx in range(len(script_names)):
                    if list(headers.keys())[0] == "package":
                        script_names[script_idx] = [script_names[script_idx][0]]
                    else:
                        script_names[script_idx] = [script_names[script_idx][1]]
            print("-" * len(f"{plugin_package_name.title()} Available Scripts"))
            print(f"{plugin_package_name.title()} Available Scripts")
            print("-" * len(f"{plugin_package_name.title()} Available Scripts"))
            if len(script_names) == 0:
                print(f"Package '{plugin_package_name.title()}' has no scripts.")
            else:
                print(
                    tabulate(
                        script_names,
                        headers=headers.values(),
                        tablefmt="rounded_grid",
                        maxcolwidths=self.terminal_width // len(headers),
                    )
                )


class GeoipsList(GeoipsCommand):
    """Top-Level List Command for listing off GeoIPS Artifacts."""

    name = "list"
    generated_classes = []
    for int_name in sorted(interfaces.__all__):
        generated_classes.append(
            CommandClassFactory(
                GeoipsListSingleInterface,
                int_name.replace("_", "-"),
            ).generated_class
        )

    command_classes = generated_classes + [
        GeoipsListInterfaces,
        GeoipsListPackages,
        GeoipsListPlugins,
        GeoipsListRegistries,
        GeoipsListScripts,
        GeoipsListSourceNames,
        GeoipsListTestDatasets,
        GeoipsListUnitTests,
    ]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Standard GeoIPS xarray dictionary based GeoKOMPSAT AMI NetCDF data reader."""

# Python Standard Libraries
from datetime import datetime, timedelta
import glob
import logging
import os

# Third-Party Libraries
import netCDF4 as ncdf
import numpy as np
import xarray

from geoips.interfaces import readers
from geoips.plugins.modules.readers.utils.geostationary_geolocation import (
    get_geolocation_cache_filename,
    get_geolocation,
)

LOG = logging.getLogger(__name__)
try:
    import numexpr as ne
except ImportError:
    LOG.info(
        "Failed import numexpr in scifile/readers/abi_ncdf4_reader_new.py. "
        "If you need it, install it."
    )

nprocs = 6

try:
    ne.set_num_threads(nprocs)
except Exception:
    LOG.info(
        f"Failed numexpr.set_num_threads in {__file__}. "
        "If numexpr is not installed and you need it, install it."
    )


interface = "readers"
family = "standard"
name = "ami_netcdf"
source_names = ["ami"]

# These should be added to the data file object
BADVALS = {
    "Off_Of_Disk": -999.9,
    "Conditional": -999.8,
    "Out_Of_Valid_Range": -999.7,
    "No_Value": -999.6,
    "Unitialized": -9999.9,
}

ALL_GEO_VARS = [
    "solar_zenith_angle",
    "satellite_zenith_angle",
    "solar_azimuth_angle",
    "satellite_azimuth_angle",
    "latitude",
    "longitude",
]

DATASET_INFO = {
    "MED": ["VI004", "VI005", "VI008"],
    "HIGH": ["VI006"],
    "LOW": [
        "NR013",
        "NR016",
        "SW038",
        "WV063",
        "WV069",
        "WV073",
        "IR087",
        "IR096",
        "IR105",
        "IR112",
        "IR123",
        "IR133",
    ],
}

CENTER_WAVENUMBERS = {
    "VI004": 0.4702,
    "VI005": 0.5086,
    "VI006": 0.6394,
    "VI008": 0.8630,
    "NR013": 1.3740,
    "NR016": 1.6092,
    "SW038": 2612.677373521110,
    "WV063": 1617.609242531340,
    "WV069": 1441.575428760170,
    "WV073": 1365.249992024440,
    "IR087": 1164.949392856340,
    "IR096": 1039.960216776110,
    "IR105": 966.153383926055,
    "IR112": 891.713057301260,
    "IR123": 810.609007871230,
    "IR133": 753.590621482278,
}


DONT_AUTOGEN_GEOLOCATION = False
if os.getenv("DONT_AUTOGEN_GEOLOCATION"):
    DONT_AUTOGEN_GEOLOCATION = True


"""
Equations and code from GEO-KOMPSAT-2A Level 1B Data User Manual.
"""


def latlon_from_lincol_geos(resolution, line, column, metadata):
    """Calculate latitude and longitude from array indices.

    Uses geostationary projection (likely won't work with extended local area files).
    Equations and code from GEO-KOMPSAT-2A Level 1B Data User Manual.
    """
    fname = get_geolocation_cache_filename("GEOLL", metadata)
    if not os.path.isfile(fname):
        degtorad = 3.14159265358979 / 180.0

        if resolution == "HIGH":
            COFF = 11000.5
            CFAC = 8.170135561335742e7
            LOFF = 11000.5
            LFAC = 8.170135561335742e7
        elif resolution == "MED":
            COFF = 5500.5
            CFAC = 4.0850677806678705e7
            LOFF = 5500.5
            LFAC = 4.0850677806678705e7
        else:
            COFF = 2750.5
            CFAC = 2.0425338903339352e7
            LOFF = 2750.5
            LFAC = 2.0425338903339352e7
        sub_lon = 128.2

        # COFF = metadata["COFF"]
        # CFAC = metadata["CFAC"]
        # LOFF = metadata["LOFF"]
        # LFAC = metadata["LFAC"]
        # sub_lon = metadata["sub_lon"]

        sub_lon = sub_lon * degtorad

        x = np.empty_like(column)
        y = np.empty_like(line)
        cosx = np.empty_like(x)
        cosy = np.empty_like(x)
        sinx = np.empty_like(x)
        siny = np.empty_like(x)
        cosxy = np.empty_like(x)
        A = np.empty_like(x)
        Sd = np.empty_like(x)
        Sn = np.empty_like(x)
        S1 = np.empty_like(x)
        S2 = np.empty_like(x)
        S3 = np.empty_like(x)
        Sxy = np.empty_like(x)

        x = degtorad * ((column - COFF) * 2**16 / CFAC)
        y = degtorad * ((line - LOFF) * 2**16 / LFAC)
        x = x.astype(np.float32)
        y = y.astype(np.float32)
        ne.evaluate("cos(x)", out=cosx)
        ne.evaluate("cos(y)", out=cosy)
        ne.evaluate("sin(x)", out=sinx)
        ne.evaluate("sin(y)", out=siny)
        ne.evaluate("cosx * cosy", out=cosxy)
        ne.evaluate("42164 * cosxy", out=A)
        B = cosy**2 + 1.006739501 * siny**2
        B = B.astype(np.float32)
        ne.evaluate("sqrt(A**2 - B*1737122264)", out=Sd)
        ne.evaluate("(A - Sd) / B", out=Sn)
        ne.evaluate("42164 - (Sn * cosxy)", out=S1)
        ne.evaluate("Sn * (sinx * cosy)", out=S2)
        ne.evaluate("-Sn * siny", out=S3)
        ne.evaluate("sqrt((S1*S1) + (S2*S2))", out=Sxy)
        nlon = (np.arctan(S2 / S1) + sub_lon) / degtorad
        nlat = np.arctan((1.006739501 * S3) / Sxy) / degtorad

        nlat[np.where(np.isnan(nlat))] = -999.9
        nlon[np.where(np.isnan(nlon))] = -999.9
        with open(fname, "w") as df:
            nlat.tofile(df)
            nlon.tofile(df)

    shape = (metadata["num_lines"], metadata["num_samples"])
    offset = 4 * metadata["num_samples"] * metadata["num_lines"]
    nlat = np.memmap(fname, mode="r", dtype=np.float32, offset=0, shape=shape)
    nlon = np.memmap(fname, mode="r", dtype=np.float32, offset=offset, shape=shape)

    return (nlat, nlon)


# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Metadata collection functions.
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
def _check_file_consistency(metadata):
    """
    Check that all input metadata are from the same image time.

    Performs cheks on satellite_name, instrument_name, observation_mode, and time.
    If these are all equal, returns True.
    If any differ, returns False.
    """
    # Checking file-level metadata for exact equality in the following fields
    checks = {
        "satellite_name": "general",
        "instrument_name": "general",
        "observation_mode": "observation",
        "observation_start_time": "observation",
    }
    for name, level in checks.items():
        check_set = set(
            [metadata[fname]["global"][level][name] for fname in metadata.keys()]
        )
        if len(check_set) != 1:
            LOG.debug("Failed on {0}. Found: {1}".format(name, check_set))
            return False
    return True


def _get_general_metadata(df):
    """Gather all of the general file-level metadata."""
    metadata = {}
    md_names = [
        "satellite_name",
        "instrument_name",
        "data_processing_center",
        "data_processing_mode",
        "channel_spatial_resolution",
        "channel_center_wavelength",
        "scene_acquisition_time",
        "mission_reference_time",
        "file_generation_time",
        "file_name",
        "file_format_version",
        "geometric_correction_sw_version",
        "star_catalog_version",
        "landmark_catalog_version",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! File-level metadata field missing: {0}".format(name))
    return metadata


def _get_data_metadata(df):
    """Gather all of the data-size metadata."""
    metadata = {}
    md_names = [
        "total_pixel_data_size",
        "number_of_total_swaths",
        "number_of_columns",
        "number_of_lines",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! Data metadata field missing: {0}".format(name))
    return metadata


def _get_observation_metadata(df):
    """Gather all of the observation metadata."""
    metadata = {}
    md_names = [
        "observation_mode",
        "observation_start_time",
        "observation_end_time",
        "time_synchro_obt",
        "time_synchro_utc",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! Observation metadata field missing: {0}".format(name))
    return metadata


def _get_projection_metadata(df):
    """Gather all of the projection metadata."""
    metadata = {}
    md_names = [
        "projection_type",
        "sub_longitude",
        "cfac",
        "lfac",
        "coff",
        "loff",
        "nominal_satellite_height",
        "earth_equatorial_radius",
        "earth_polar_radius",
        "earth_ellipsoid_center",
        "image_upperleft_latitude",
        "image_upperleft_longitude",
        "image_lowerright_latitude",
        "image_lowerright_longitude",
        "image_center_latitude",
        "image_center_longitude",
        "image_upperleft_x",
        "image_upperleft_y",
        "image_lowerright_x",
        "image_lowerright_y",
        "resampling_kernel_type",
    ]

    for name in md_names:
        if hasattr(df, name):
            metadata[name] = getattr(df, name)
        else:
            LOG.info("Warning! Projection metadata field missing: {0}".format(name))
    return metadata


def _get_global_metadata(df):
    """Get file-level metadata."""
    metadata = {}
    metadata["general"] = _get_general_metadata(df)
    metadata["data"] = _get_data_metadata(df)
    metadata["observation"] = _get_observation_metadata(df)
    metadata["projection"] = _get_projection_metadata(df)
    return metadata


def _get_variable_metadata(df):
    """Gather all of the variable metadata."""
    metadata = {}
    md_names = [
        "channel_name",
        "detector_side",
        "number_of_total_pixels",
        "number_of_error_pixels",
        "max_pixel_value",
        "max_pixel_value",
        "average_pixel_value",
        "stddev_pixel_value",
        "number_of_total_bits_per_pixel",
        "number_of_data_quality_flag_bits_per_pixel",
        "number_of_valid_bits_per_pixel",
        "data_quality_flag_meaning",
        "ground_sample_distance_ew",
        "ground_sample_distance_ns",
    ]

    for name in md_names:
        if hasattr(df.variables["image_pixel_values"], name):
            metadata[name] = getattr(df.variables["image_pixel_values"], name)
        else:
            LOG.info("Warning! Projection metadata field missing: {0}".format(name))
    return metadata


def _get_metadata(df, fname):
    """
    Gather metadata for the data file and return as a dictionary.

    Note: We are gathering all of the available metadata in case it is needed at
    some point.
    """
    metadata = {}
    # Gather all file-level metadata.
    metadata["global"] = _get_global_metadata(df)
    # Gather all variable-level metadata.
    metadata["var_info"] = _get_variable_metadata(df)
    # Gather some useful info to the top level.
    try:
        metadata["path"] = df.filepath()
    except ValueError:
        # Without cython installed, df.filepath() does not work
        metadata["path"] = fname
    metadata["satellite"] = metadata["global"]["general"]["satellite_name"]
    metadata["sensor"] = df.instrument_name
    # These were being returned as np.int32's, causing an overflow when passed
    # to numpy 2.x memmap call in latlon_from_lincol_geos function below.
    # Explicitly cast to np.int64 to avoid overflow in memmap
    metadata["num_lines"] = np.int64(df.number_of_lines)
    metadata["num_samples"] = np.int64(df.number_of_columns)
    return metadata


def metadata_to_datetime(metadata):
    """Use information from the metadata to get the image datetime."""
    start_time = metadata["global"]["observation"]["observation_start_time"]
    end_time = metadata["global"]["observation"]["observation_end_time"]
    epoch = datetime(2000, 1, 1, 12, 0, 0)
    start_time = epoch + timedelta(seconds=start_time)
    end_time = epoch + timedelta(seconds=end_time)
    return start_time, end_time


def _get_geolocation_metadata(metadata):
    """
    Gather all of the metadata used in creating geolocation data for input file.

    This is split out so we can easily create a chash of the data for creation
    of a unique filename. This allows us to avoid recalculation of angles that
    have already been calculated.
    """
    metadata = metadata["global"]
    geomet = {}
    geomet["platform_name"] = metadata["general"]["satellite_name"]
    geomet["Re"] = metadata["projection"]["earth_equatorial_radius"]
    geomet["Rp"] = metadata["projection"]["earth_polar_radius"]
    geomet["e"] = 0.0818191910435
    # Nominal satellite height is from the center of the Earth.
    geomet["H_m"] = metadata["projection"]["nominal_satellite_height"]
    geomet["pphgt"] = geomet["H_m"] - geomet["Re"]
    geomet["lat0"] = metadata["projection"]["image_center_latitude"]
    geomet["lon0"] = metadata["projection"]["image_center_longitude"] * 180.0 / np.pi
    geomet["scene"] = metadata["observation"]["observation_mode"]
    # Just getting the nadir resolution in kilometers.  Must extract from a string.
    geomet["res_km"] = float(metadata["general"]["channel_spatial_resolution"])
    geomet["roi_factor"] = 5  # roi = res * roi_factor, was 10
    # These were being returned as np.int32's, causing an overflow when passed
    # to numpy 2.x memmap call in latlon_from_lincol_geos function below.
    # Explicitly cast to np.int64 to avoid overflow in memmap
    geomet["num_lines"] = np.int64(metadata["data"]["number_of_lines"])
    geomet["num_samples"] = np.int64(metadata["data"]["number_of_columns"])
    # Dynamically get offsets and scale factors needed for correct geolocation
    # calculation
    geomet["sub_lon"] = metadata["projection"]["sub_longitude"]
    geomet["CFAC"] = np.abs(metadata["projection"]["cfac"])
    geomet["COFF"] = np.abs(metadata["projection"]["coff"])
    geomet["LFAC"] = np.abs(metadata["projection"]["lfac"])
    geomet["LOFF"] = np.abs(metadata["projection"]["loff"])
    return geomet


# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# Get pixel data.
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
def get_data(gvars, fname, rad=False, ref=False, bt=False):
    """Get actual pixel data from a specific file using geolocation information."""
    if "Lines" in gvars and "Samples" in gvars:
        full_disk = False
        line_inds = gvars["Lines"]
        sample_inds = gvars["Samples"]
    else:
        full_disk = True

    ds = ncdf.Dataset(fname, "r")
    ipixel = ds.variables["image_pixel_values"]

    if not full_disk:
        counts = ipixel[:].data[line_inds, sample_inds]
    else:
        # Leaving out zoom stuff in abi reader for now, since it seems
        #   those conditions shouldn't be reached for these files.
        counts = ipixel[:].data

    # The following section of bitwise "masking" is from the manual.
    # I believe what is expected to happen is that, if a pixel is out-of-area
    # or erroneous, it will be exactly 32768 or 49152, and will be 0 after the
    # masking. Conditional pixels will be treated as "good enough" and will have
    # a value less than 16384 (i.e. the "normal data" range after the bitwise
    # mask is applied.
    channel = ipixel.getncattr("channel_name")
    if (channel == "VI004") or (channel == "VI005") or (channel == "NR016"):
        mask = 0b0000011111111111  # 11bit mask
    elif (channel == "VI006") or (channel == "NR013") or (channel == "WV063"):
        mask = 0b0000111111111111  # 12bit mask
    elif channel == "SW038":
        mask = 0b0011111111111111  # 14bit mask
    else:
        mask = 0b0001111111111111  # 13bit mask

    if channel.startswith("VI") or channel.startswith("NR"):
        reflectance_chan = True
    else:
        reflectance_chan = False

    counts_masked = np.bitwise_and(counts, mask)

    # So far, looks like the binary masking from the manual is sufficient.
    # (Good and conditional values will be kept.)
    # The following code is a starting point if conditional values need to be
    # masked.
    # conditional_threshold = 16384
    # out_of_area_threshold = 32768
    # erroneous_threshold = 49152
    # counts_masked = np.ma.masked_where(counts >= conditional_threshold, counts)

    gain = ds.DN_to_Radiance_Gain
    offset = ds.DN_to_Radiance_Offset

    # Equations from last page of manual (Table 14)
    # radiance = gain * count + offset
    radiance = gain * counts_masked + offset

    data = {}
    if rad:
        data["Rad"] = radiance

    if reflectance_chan and ref:
        # Visible and Near IR
        c_prime = ds.Radiance_to_Albedo_c
        albedo = radiance * c_prime
        data["Ref"] = albedo

    if bt and not reflectance_chan:
        # IR
        # Using hard-coded values for nu, since the values obtained from the
        #   dataset don't quite match the values from the manual.
        nu = CENTER_WAVENUMBERS[channel]
        h = ds.Plank_constant_h
        k = ds.Boltzmann_constant_k
        c = ds.light_speed
        c0 = ds.Teff_to_Tbb_c0
        c1 = ds.Teff_to_Tbb_c1
        c2 = ds.Teff_to_Tbb_c2
        T_e = (h * c / k * (nu * 100)) / np.log(
            (2 * h * c**2) * (nu * 100) ** 3 / (radiance * 1e-5) + 1
        )
        T_b = c0 + c1 * T_e + c2 * T_e**2

        data["BT"] = T_b

    ds.close()

    return data


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """
    Read Geo-Kompsat NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    return readers.read_data_to_xarray_dict(
        fnames,
        call_single_time,
        metadata_only,
        chans,
        area_def,
        self_register,
    )


def call_single_time(
    fnames, metadata_only=False, chans=None, area_def=None, self_register=False
):
    """
    Read Geo-Kompsat NetCDF data from a list of filenames.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * Return before actually reading data if True
    chans : list of str, default=None
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    gvars = {}
    datavars = {}
    geo_metadata = {}

    # Get metadata for all input data files
    # Check to be sure that all input files are form the same image time
    all_metadata = {}
    for fname in fnames:
        if chans:
            gotone = False
            for chan in chans:
                currchan = chan[:5].lower()
                if currchan in fname:
                    gotone = True
            if not gotone:
                LOG.info(
                    "SKIPPING file %s, not needed from channel list %s", fname, chans
                )
                continue
        try:
            # Open using with to avoid seg faults due to not properly closing files.
            # This did not seg fault prior to netcdf 1.7 / numpy 2.x
            with ncdf.Dataset(str(fname), "r") as ds:
                all_metadata[fname] = _get_metadata(ds, fname)
        except IOError as resp:
            LOG.exception("BAD FILE %s skipping", resp)
            continue
        if metadata_only:
            LOG.info("Only need metadata from first file, returning")
            break

    if not _check_file_consistency(all_metadata):
        raise ValueError("Input files inconsistent.")

    # Now put together a dict that shows what we actually got
    # This is largely to help us read in order and only create arrays of the minimum
    # required size.
    # Dict structure is channel{metadata}
    file_info = {}

    xarray_obj = xarray.Dataset()
    for fname, md in all_metadata.items():
        ch = md["var_info"]["channel_name"]
        if ch not in file_info:
            file_info[ch] = md

    xarray_obj.attrs["file_metadata"] = file_info.copy()

    # Most of the metadata are the same between files.
    # From here on we will just rely on the metadata from a single data file
    # for each resolution
    res_md = {}
    for res in ["LOW", "MED", "HIGH"]:
        # Find a file for this resolution: Any one will do
        res_chans = list(set(DATASET_INFO[res]).intersection(file_info.keys()))
        if res_chans:
            res_md[res] = file_info[res_chans[0]]

    # If we plan to self register, make sure we requested a resolution that we actually
    # plan to read.
    #
    # This could be problematic if we try to self-register to LOW when only
    # reading MED or something
    if self_register and self_register not in res_md:
        raise ValueError(
            "Resolution requested for self registration has not been read."
        )

    # Gather metadata
    # Assume the same for all resolutions.  Not true, but close enough.
    highest_md = res_md[list(res_md.keys())[0]]
    start_dt, end_dt = metadata_to_datetime(highest_md)
    xarray_obj.attrs["start_datetime"] = start_dt
    xarray_obj.attrs["end_datetime"] = end_dt
    xarray_obj.attrs["source_name"] = "ami"  # "ami"
    xarray_obj.attrs["data_provider"] = "nmsc"
    xarray_obj.attrs["platform_name"] = highest_md["global"]["general"][
        "satellite_name"
    ]
    xarray_obj.attrs["area_definition"] = area_def
    xarray_obj.attrs["source_file_names"] = [
        os.path.basename(fname) for fname in fnames
    ]

    # Get appropriate sector name
    if area_def:
        xarray_obj.attrs["area_id"] = area_def.area_id
        xarray_obj.attrs["area_definition"] = area_def
    else:
        if self_register and self_register not in DATASET_INFO:
            raise ValueError(
                f"Unrecognized resolution name requested for self "
                f"registration: {self_register}"
            )
        scene_id = highest_md["global"]["observation"]["observation_mode"].lower()
        if scene_id == "fd":
            xarray_obj.attrs["area_id"] = "Full-Disk"
        elif scene_id == "ela":
            xarray_obj.attrs["area_id"] = "Extended-Local-Area"
        elif scene_id == "la":
            xarray_obj.attrs["area_id"] = "Local-Area"
        elif scene_id == "enh":
            xarray_obj.attrs["area_id"] = "Extend-Northern-Hemisphere"
        elif scene_id == "ea":
            xarray_obj.attrs["area_id"] = "East-Asia"
        elif scene_id == "ko":
            xarray_obj.attrs["area_id"] = "Korea"
        elif scene_id == "tp":
            xarray_obj.attrs["area_id"] = "Typhoon"
        xarray_obj.attrs["area_definition"] = None
    adname = xarray_obj.attrs["area_id"]

    if metadata_only:
        LOG.info("Only need metadata from first file, returning")
        return {"METADATA": xarray_obj}

    # Create list of all possible channels for the case where no channels were requested
    all_chans_list = []
    for chan_list in DATASET_INFO.values():
        for chl in chan_list:
            all_chans_list.append(chl + "Rad")
            if chl.startswith("VI") or chl.startswith("NR"):
                all_chans_list.append(chl + "Ref")
            else:
                all_chans_list.append(chl + "BT")

    # If specific channels were requested, check them against the input data
    # If specific channels were requested, but no files exist for one of the
    # channels, then error
    if chans:
        for chan in chans:
            if chan in ALL_GEO_VARS:
                continue
            if chan not in all_chans_list:
                raise ValueError("Requested channel {0} not recognized.".format(chan))
            if chan[0:5] not in file_info.keys():
                continue

    # If no specific channels were requested, get everything
    if not chans:
        chans = all_chans_list

    # Creates a dictionary whose keys are band numbers in the form {name}##
    # and whose values are lists containing the data types(s) requested for
    # the band (e.g. Rad, Ref, BT).
    chan_info = {}
    for ch in chans:
        if ch in ALL_GEO_VARS:
            continue
        chn = ch[0:5]
        typ = ch[5:]
        if chn not in chan_info:
            chan_info[chn] = []
        chan_info[chn].append(typ)

    # Gather geolocation data
    # Assume datetime the same for all resolutions.  Not true, but close enough.
    # This saves us from having very slightly different solar angles for each channel.
    # Loop over resolutions and get metadata as needed
    if self_register:
        LOG.info("")
        LOG.info("Getting geolocation information for adname %s.", adname)
        geo_metadata[adname] = _get_geolocation_metadata(res_md[self_register])

        i = np.arange(0, geo_metadata[adname]["num_lines"], dtype="f")
        j = np.arange(0, geo_metadata[adname]["num_samples"], dtype="f")
        i, j = np.meshgrid(i, j)
        (fldk_lats, fldk_lons) = latlon_from_lincol_geos(
            resolution=self_register, column=i, line=j, metadata=geo_metadata[adname]
        )

        gvars[adname] = get_geolocation(
            start_dt, geo_metadata[adname], fldk_lats, fldk_lons, BADVALS, area_def
        )
        if not gvars[adname]:
            LOG.error(
                "GEOLOCATION FAILED for adname %s DONT_AUTOGEN_GEOLOCATION is: %s",
                adname,
                DONT_AUTOGEN_GEOLOCATION,
            )
            gvars[adname] = {}
    else:
        for res in ["LOW", "MED", "HIGH"]:
            try:
                res_md[res]
            except KeyError:
                continue
            LOG.info("")
            LOG.info(
                "Getting geolocation information for resolution %s for %s", res, adname
            )
            try:
                geo_metadata[res] = _get_geolocation_metadata(res_md[res])

                i = np.arange(0, geo_metadata[res]["num_lines"], dtype="f")
                j = np.arange(0, geo_metadata[res]["num_samples"], dtype="f")
                i, j = np.meshgrid(i, j)
                (fldk_lats, fldk_lons) = latlon_from_lincol_geos(
                    resolution=res, column=i, line=j, metadata=geo_metadata[res]
                )

                gvars[res] = get_geolocation(
                    start_dt, geo_metadata[res], fldk_lats, fldk_lons, BADVALS, area_def
                )
            except IndexError as resp:
                LOG.exception("SKIPPING apparently no coverage or bad geolocation file")
                raise IndexError(resp)

    LOG.interactive("Done with geolocation for {}".format(adname))
    LOG.info("")

    # Read the data
    # Will read all data if sector_definition is None
    # Will only read required data if an sector_definition is provided
    for chan, types in chan_info.items():
        # If we didn't pass all the channel data, skip non-existent data types
        if chan not in file_info:
            continue
        LOG.info("Reading {}".format(chan))
        chan_md = file_info[chan]
        for res, res_chans in DATASET_INFO.items():
            if chan in res_chans:
                break
        if (not self_register) and (
            res not in gvars.keys() or not gvars[res] or "latitude" not in gvars[res]
        ):
            LOG.info(
                f"We don't have geolocation information for {res} for {adname} "
                f"skipping {chan}."
            )
            continue
        if not area_def:
            dsname = res
        else:
            dsname = adname

        if self_register:
            gname = adname
        else:
            gname = res

        rad = ref = bt = False
        if "Rad" in types:
            rad = True
        if "Ref" in types:
            ref = True
        if "BT" in types:
            bt = True

        data = get_data(gvars[gname], chan_md["path"], rad, ref, bt)
        for typ, val in data.items():
            if dsname not in datavars:
                datavars[dsname] = {}
            datavars[dsname][chan + typ] = val

    if area_def:
        for res in ["LOW", "MED", "HIGH"]:
            if res in gvars and gvars[res]:
                gvars_res = gvars.pop(res)
                if adname not in gvars:
                    gvars[adname] = gvars_res

    if self_register:
        # Determine which resolution has geolocation
        LOG.info("Registering to {}".format(self_register))
        all_res = ["LOW", "MED", "HIGH"]
        if self_register not in all_res:
            raise ValueError("No geolocation data found.")

        all_res.remove(self_register)
        datavars[adname] = datavars.pop(self_register)
        for res in all_res:
            if res in datavars:
                for varname, var in datavars[res].items():
                    datavars[adname][varname] = var
                datavars.pop(res)

    # Remove lines and samples arrays.  Not needed.
    for res in gvars.keys():
        try:
            gvars[res].pop("Lines")
            gvars[res].pop("Samples")
            for varname, var in gvars[res].items():
                gvars[res][varname] = np.ma.array(
                    var, mask=gvars[res]["satellite_zenith_angle"].mask
                )
                gvars[res][varname] = np.ma.masked_where(
                    gvars[res]["satellite_zenith_angle"] > 75, gvars[res][varname]
                )
        except KeyError:
            pass
    for ds in datavars.keys():
        if not datavars[ds]:
            datavars.pop(ds)

    # Create the final dictionary of xarray objects.
    xarray_objs = {}
    for dsname in datavars.keys():
        xobj = xarray.Dataset()
        xobj.attrs = xarray_obj.attrs.copy()
        for varname in datavars[dsname].keys():
            xobj[varname] = xarray.DataArray(datavars[dsname][varname])
        for varname in gvars[dsname].keys():
            xobj[varname] = xarray.DataArray(gvars[dsname][varname])

        roi = 500
        if hasattr(xobj, "area_definition") and xobj.area_definition is not None:
            roi = max(
                xobj.area_definition.pixel_size_x, xobj.area_definition.pixel_size_y
            )
            LOG.info("Trying area_def roi %s", roi)
        for curr_res in geo_metadata.keys():
            LOG.info(
                "Trying metadata roi %s %s",
                geo_metadata[curr_res]["res_km"] * 1000.0,
                roi,
            )
            if geo_metadata[curr_res]["res_km"] * 1000.0 > roi:
                roi = geo_metadata[curr_res]["res_km"] * 1000.0
                LOG.info("Trying standard_metadata[%s] %s", curr_res, roi)
        xobj.attrs["interpolation_radius_of_influence"] = roi
        xarray_objs[dsname] = xobj
        # At some point we may need to deconflict, but for now just use any of the
        # dataset attributes as the METADATA dataset
        xarray_objs["METADATA"] = xobj[[]]

    LOG.info("Done reading GEOKOMPSAT AMI data for %s", adname)
    LOG.info("")

    return xarray_objs


# Unit test functions
def get_test_files(test_data_dir):
    """Generate testing xarray from test data."""
    filepath = test_data_dir + "/test_data_noaa_aws/data/geokompsat/20231208/0300/*.nc"
    filelist = glob.glob(filepath)
    tmp_xr = call(filelist)
    if len(filelist) == 0:
        raise NameError("No files found")
    return tmp_xr


def get_test_parameters():
    """Generate test data key for unit testing."""
    return [
        {"data_key": "HIGH", "data_var": "VI006Ref", "mean": 0.14788916},
        {"data_key": "LOW", "data_var": "IR112BT", "mean": 286.63754873},
    ]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""GeoIPS decorators module."""

from warnings import warn
from functools import wraps
import inspect


class deprecated:
    """
    A decorator that deprecates a function.

    When applied to a function, will cause that function to raise a
    DeprecationWarning when called.
    """

    def __init__(self, replacement=None):
        self.replacement = replacement

    def __call__(self, func):
        """Call method."""
        func_mod = inspect.getmodule(func).__name__
        func_name = f"{func_mod}.{func.__name__}"
        msg = f"Function '{func_name}' is deprecated and "
        msg += "will be removed in a future version."
        if self.replacement:
            repl_mod = inspect.getmodule(self.replacement).__name__
            repl_name = f"{repl_mod}.{self.replacement.__name__}"
            msg += f' Use "{repl_name}" for the same functionality.'

        @wraps(func)
        def wrapped_func(*args, **kwargs):
            warn(msg, DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)

        wrapped_func.__doc__ = f"DECPRECATED: {msg}\n\n"
        wrapped_func.__doc__ += self.replacement.__doc__
        return wrapped_func


def developmental(func):
    """Mark an interfaces as developmental.

    When applied to a function, will prepend a "developmental" message to the
    beginning of that function's docstring.
    """
    func_mod = inspect.getmodule(func).__name__
    func_name = f"{func_mod}.{func.__name__}"
    msg = f"'{func_name}' is currently under development "
    msg += "and should be expected to change in the future.\n"
    msg += "Please provide feedback at "
    msg += "https://github.com/NRLMMD-GEOIPS/geoips or geoips@nrlmry.navy.mil."

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        return func(*args, **kwargs)

    wrapped_func.__doc__ = f"DEVELOPMENTAL: {msg}\n\n"
    wrapped_func.__doc__ += func.__doc__

    return wrapped_func



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Filename formatter for text windspeed products."""

import logging

from os.path import join as pathjoin

from geoips.filenames.base_paths import PATHS as gpaths

LOG = logging.getLogger(__name__)

interface = "filename_formatters"
family = "xarray_metadata_to_filename"
name = "text_winds_full_fname"


def call(
    xarray_obj,
    extension=".txt",
    basedir=pathjoin(gpaths["ANNOTATED_IMAGERY_PATH"], "text_winds"),
):
    """Create a single text winds file for all data in the current xarray.

    This text windspeed filename includes YYYYMMDD.HHMN in the filename
    in order to include only the current datafile in the file.

    See Also
    --------
    geoips.plugins.modules.filename_formatters.text_winds_full_fname.
        assemble_windspeeds_text_full_fname
        Shared utility to create filenames with similar formatting requirements.
    """
    return assemble_windspeeds_text_full_fname(
        basedir=basedir,
        source_name=xarray_obj.source_name,
        platform_name=xarray_obj.platform_name,
        data_provider=xarray_obj.data_provider,
        product_datetime=xarray_obj.start_datetime,
        dt_format="%Y%m%d.%H%M",
        extension=extension,
        creation_time=None,
    )


def assemble_windspeeds_text_full_fname(
    basedir,
    source_name,
    platform_name,
    data_provider,
    product_datetime,
    dt_format="%Y%m%d.%H%M",
    extension=".txt",
    creation_time=None,
):
    """Produce full output product path using product / sensor specifications.

    Parameters
    ----------
    basedir : str
         base directory
    source_name : str
        Name of source (sensor)
    platform_name : str
        Name of platform (satellite)
    data_provider : str
        Name of data provider
    product_datetime : datetime.datetime
        Start time of data used to generate product
    dt_format : str, default="%Y%m%d.%H%M"
        Format used to display product_datetime within filename
    extension : str, default=".txt"
        File extension, specifying type.
    creation_time : datetime.datetime, default=None
        Include given creation_time of file in filename
        If None, do not include creation time.

    Returns
    -------
    str
        full path of output filename of the format:
          <basedir>/<source_name>_<data_provider>_<platform_name>_
          surface_winds_<YYYYMMDDHHMN>

    Examples
    --------
    >>> startdt = datetime.strptime('20200216T001412', '%Y%m%dT%H%M%S')
    >>> assemble_windspeeds_text_full_fname(
    ...     '/outdir',
    ...     'smap-spd',
    ...     'smap',
    ...     'remss',
    ...     startdt,
    ...     '%Y%m%d'
    ...     )
    '/outdir/smap-spd_remss_smap_surface_winds_20200216'
    """
    fname = "_".join(
        [
            source_name,
            data_provider,
            platform_name,
            "surface_winds",
            product_datetime.strftime(dt_format),
        ]
    )

    if creation_time is not None:
        fname = fname + "_creationtime_" + creation_time.strftime("%Y%m%dT%H%MZ")

    if extension is not None:
        fname = fname + extension

    return pathjoin(basedir, fname)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Matplotlib-based windbarb annotated image output."""

import logging

import numpy

from geoips.image_utils.mpl_utils import (
    create_figure_and_main_ax_and_mapobj,
    save_image,
)
from geoips.image_utils.colormap_utils import set_matplotlib_colors_standard
from geoips.image_utils.mpl_utils import plot_image, plot_overlays, create_colorbar
from geoips.image_utils.mpl_utils import get_title_string_from_objects, set_title
from geoips.plugins.modules.output_formatters.imagery_windbarbs import (
    format_windbarb_data,
    output_clean_windbarbs,
    plot_barbs,
)

LOG = logging.getLogger(__name__)

interface = "output_formatters"
family = "image_overlay"
name = "imagery_windbarbs_multi_level"


def assign_height_levels(windbarb_data_dict, pressure_range_dict):
    """Assing derived motion winds to specified height levels.

    Using the pressure associated with each retrieved wind observation, assign to
    a specified level (e.g. Low/Mid/High) based on predefined pressure ranges. Each
    pressure level is assigned an integer, and any unassigned are set to 0

    Parameters
    ----------
    formatted_data_dict : dict
        Dictionary holding DMW data - must include a pressure key
    pressure_range_dict : dict
        Dictionary specifying pressure range for each defined level.
        e.g. {"Low": [701, 1013.25], "Mid": [401, 700], "High": [0, 400]}

    Returns
    -------
    tuple
        Array of assigned level numbers, and list of associated labels that can be used
        to set the ticks on a colorbar
    """
    pressure = windbarb_data_dict["pressure"]
    n_valid = numpy.count_nonzero(pressure)
    height_num = numpy.zeros(pressure.shape)
    level_labels = ["Unassigned"]
    for i, (level, pressure_range) in enumerate(pressure_range_dict.items()):
        max_pres = max(pressure_range)
        min_pres = min(pressure_range)
        pressure_mask = (pressure.data >= min_pres) & (pressure.data <= max_pres)
        height_num[pressure_mask] = i + 1
        level_labels.append(f"{level}\n({max_pres} - {min_pres} hPa)")
        LOG.info(
            "Number of wind retrievals for %s: %s",
            level,
            numpy.count_nonzero(pressure_mask),
        )
    LOG.info("Assigned %s/%s retrievals", numpy.count_nonzero(height_num), n_valid)
    return height_num, level_labels


def call(
    area_def,
    xarray_obj,
    product_name,
    output_fnames,
    clean_fname=None,
    product_name_title=None,
    mpl_colors_info=None,
    feature_annotator=None,
    gridline_annotator=None,
    product_datatype_title=None,
    bg_data=None,
    bg_mpl_colors_info=None,
    bg_xarray=None,
    bg_product_name_title=None,
    bg_datatype_title=None,
    remove_duplicate_minrange=None,
    title_copyright=None,
    title_formatter=None,
    pressure_range_dict=None,
):
    """Plot annotated windbarbs on matplotlib figure."""
    LOG.info("Starting imagery_windbarbs")

    if product_name_title is None:
        product_name_title = product_name

    success_outputs = []

    # Plot windbarbs

    formatted_data_dict = format_windbarb_data(xarray_obj, product_name)

    if not pressure_range_dict:
        pressure_range_dict = {
            "Low": [701, 1013.25],
            "Mid": [401, 700],
            "High": [0, 400],
        }
    height_numbers, level_labels = assign_height_levels(
        formatted_data_dict, pressure_range_dict
    )
    formatted_data_dict["height_numbers"] = height_numbers

    if clean_fname is not None:
        success_outputs += output_clean_windbarbs(
            area_def,
            [clean_fname],
            mpl_colors_info,
            xarray_obj.start_datetime,
            formatted_data_dict,
            barb_color_variable="height_numbers",
        )

    if output_fnames is not None:
        LOG.info("Starting output_fnames")

        # Create matplotlib figure and main axis, where the main image will be plotted
        fig, main_ax, mapobj = create_figure_and_main_ax_and_mapobj(
            area_def.x_size,
            area_def.y_size,
            area_def,
            existing_mapobj=None,
            noborder=False,
        )

        if bg_data is not None:
            if not bg_mpl_colors_info:
                bg_mpl_colors_info = set_matplotlib_colors_standard(
                    data_range=[bg_data.min(), bg_data.max()],
                    cmap_name="Greys",
                    cbar_label=None,
                    create_colorbar=False,
                )
            # Plot the background data on a map
            plot_image(main_ax, bg_data, mapobj, mpl_colors_info=bg_mpl_colors_info)

        plot_barbs(
            main_ax,
            mapobj,
            mpl_colors_info,
            formatted_data_dict,
            barb_color_variable="height_numbers",
        )

        # Set the title for final image
        title_string = get_title_string_from_objects(
            area_def,
            xarray_obj,
            product_name_title,
            product_datatype_title=product_datatype_title,
            bg_xarray=bg_xarray,
            bg_product_name_title=bg_product_name_title,
            bg_datatype_title=bg_datatype_title,
            title_copyright=title_copyright,
            title_formatter=title_formatter,
        )
        set_title(main_ax, title_string, area_def.y_size)

        if mpl_colors_info["colorbar"] is True:
            mpl_colors_info["cbar_tick_labels"] = level_labels
            mpl_colors_info["colorbar_kwargs"] = {"extend": "neither"}
            # Create the colorbar to match the mpl_colors
            create_colorbar(fig, mpl_colors_info)

        # Plot gridlines and feature overlays
        plot_overlays(
            mapobj,
            main_ax,
            area_def,
            feature_annotator=feature_annotator,
            gridline_annotator=gridline_annotator,
        )

        for annotated_fname in output_fnames:
            # Save the final image
            success_outputs += save_image(
                fig,
                annotated_fname,
                is_final=True,
                image_datetime=xarray_obj.start_datetime,
            )

    return success_outputs



# # # This source code is protected under the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Fixtures for testing the Order-based procflow pydantic models."""

# Third-Party Libraries
import pytest


@pytest.fixture
def valid_step_data():
    """Fixture to provide sample valid plugin data for testing."""
    return {
        "kind": "reader",
        "name": "abi_netcdf",
        "arguments": {
            "area_def": "None",
            "variables": ["None"],
            "metadata_only": False,
            "self_register": ["None"],
        },
    }


@pytest.fixture
def valid_interfaces(valid_plugin_kinds):
    """Fixture to provide list of valid GeoIPS interfaces."""
    return {f"{plugin_kind}s" for plugin_kind in valid_plugin_kinds}


@pytest.fixture
def valid_reader_arguments_model_data():
    """Fixture to provide sample valid Reader arguments for testing."""
    return {
        "area_def": "None",
        "variables": ["None"],
        "metadata_only": True,
        "self_register": ["None"],
        "fnames": ["None"],
    }


@pytest.fixture
def valid_plugin_kinds():
    """Fixture to provide the list of valid plugin kinds."""
    VALID_PLUGIN_KINDS = [
        "algorithm",
        "procflow",
        "database",
        "sector_adjuster",
        "output_checker",
        "output_formatter",
        "reader",
        "gridline_annotator",
        "product_default",
        "sector_metadata_generator",
        "product",
        "sector_spec_generator",
        "title_formatter",
        "coverage_checker",
        "feature_annotator",
        "colormapper",
        "sector",
        "interpolator",
        "filename_formatter",
        "workflow",
    ]
    return VALID_PLUGIN_KINDS


# test_bases.py
@pytest.fixture
def valid_plugin_data():
    """Fixture providing valid sample data for Plugin model."""
    return {
        "interface": "workflows",
        "family": "geoips_family",
        "name": "read_test",
        "docstring": "This is a valid numpy docstring.",
        "description": "This is a valid numpy docstring.",
        "package": "geoips",
        "relpath": "geoips/tests/unit_tests/pydantic",
        "abspath": "/home/kumar/geoips/geoips/tests/unit_tests/pydantic",
    }


@pytest.fixture
def pluign_types_and_plugins():
    """Fixture providing valid plugin types & corresponding plugin names."""
    return {
        "reader": [
            "scat_knmi_winds_netcdf",
            "abi_netcdf",
            "mimic_netcdf",
            "geoips_netcdf",
            "seviri_hrit",
            "amsub_mirs",
            "atms_hdf5",
            "amsr2_remss_winds_netcdf",
            "modis_hdf4",
            "wfabba_ascii",
            "saphir_hdf5",
            "ssmis_binary",
            "ewsg_netcdf",
            "viirs_sdr_hdf5",
            "amsr2_netcdf",
            "amsub_hdf",
            "gmi_hdf5",
            "viirs_netcdf",
            "windsat_remss_winds_netcdf",
            "scat_noaa_winds_netcdf",
            "windsat_idr37_binary",
            "smos_winds_netcdf",
            "imerg_hdf5",
            "ahi_hsd",
            "ascat_uhr_netcdf",
            "sar_winds_netcdf",
            "ami_netcdf",
            "ssmi_binary",
            "cygnss_netcdf",
            "abi_l2_netcdf",
            "sfc_winds_text",
            "smap_remss_winds_netcdf",
            "clavrx_netcdf4",
            "clavrx_hdf4",
        ],
    }



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Read NASA GPM GMI hdf5 data files.

Read a grannual NASA GPM GMI TBs in h5 format (each grannual file is about
5 minutes GPM GMI data)

Output variables in xarray object for geoips processing system

V0:   August 4, 2020

Dataset information::

    variables in original TBs structure format
    tb_info = { 'S1': {  'tb10v': 0,
                         'tb10h': 1,
                         'tb19v': 2,
                         'tb19h': 3,
                         'tb23v': 4,
                         'tb37v': 5,
                         'tb37h': 6,
                         'tb89v': 7,
                         'tb89h': 8,},
                 'S2': { 'tb166v': 0,
                         'tb166h': 1,
                         'tb183_3v': 2,
                         'tb183_7v': 3}
             }
"""
# Python Standard Libraries
from glob import glob
import logging
from os.path import basename

# Third-Party Libraries
import h5py
import matplotlib
import numpy as np
import pandas as pd
import xarray as xr


matplotlib.use("agg")
LOG = logging.getLogger(__name__)

interface = "readers"
family = "standard"
name = "gmi_hdf5"
source_names = ["gmi"]


def read_gmi_file(fname, xarray_gmi):
    """Read a single GMI file fname."""
    fileobj = h5py.File(fname, mode="r")
    # get the variables ( tbt/lon(nscan,npix), tb(nscan,npix,nChan),....., time(ns))

    lon = fileobj["S1"]["Longitude"][()]
    lat = fileobj["S1"]["Latitude"][()]
    tb = fileobj["S1"]["Tb"][()]
    tb_hi = fileobj["S2"]["Tb"][()]  # for 166 and 183-7 GHz

    # time info for each scan
    yy = fileobj["S1"]["ScanTime"]["Year"][()]
    mo = fileobj["S1"]["ScanTime"]["Month"][()]
    dd = fileobj["S1"]["ScanTime"]["DayOfMonth"][()]
    hh = fileobj["S1"]["ScanTime"]["Hour"][()]
    mm = fileobj["S1"]["ScanTime"]["Minute"][()]
    ss = fileobj["S1"]["ScanTime"]["Second"][()]

    # setup time in datetime64 format required by geoips

    nscan = lat.shape[0]
    npix = lat.shape[1]  # 221 pixels per scan
    time_scan = np.zeros((nscan, npix))

    for i in range(nscan):
        time_scan[i:] = "%04d%02d%02d%02d%02d%02d" % (
            yy[i],
            mo[i],
            dd[i],
            hh[i],
            mm[i],
            ss[i],
        )

    # assignment of TB at each channel
    V10 = tb[:, :, 0]
    H10 = tb[:, :, 1]
    V19 = tb[:, :, 2]
    H19 = tb[:, :, 3]
    V23 = tb[:, :, 4]
    V37 = tb[:, :, 5]
    H37 = tb[:, :, 6]
    V89 = tb[:, :, 7]
    H89 = tb[:, :, 8]

    V166 = tb_hi[:, :, 0]
    H166 = tb_hi[:, :, 1]
    V183_3 = tb_hi[:, :, 2]
    V183_7 = tb_hi[:, :, 3]

    # close the h5 object
    fileobj.close()

    #          ------  setup xarray variables   ------

    # namelist_gmi  = ['latitude', 'longitude', 'V10', 'H10', 'V19','H19','V23', 'V37',
    #                  'H37', 'V89' ,'H89', 'V166', 'H166', 'V183-3','V183-7', 'time']

    final_xarray = xr.Dataset()
    if "latitude" not in xarray_gmi.variables.keys():
        # setup GMI xarray
        final_xarray["latitude"] = xr.DataArray(lat)
        final_xarray["longitude"] = xr.DataArray(lon)
        final_xarray["V10"] = xr.DataArray(V10)
        final_xarray["H10"] = xr.DataArray(H10)
        final_xarray["V19"] = xr.DataArray(V19)
        final_xarray["H19"] = xr.DataArray(H19)
        final_xarray["V23"] = xr.DataArray(V23)
        final_xarray["V37"] = xr.DataArray(V37)
        final_xarray["H37"] = xr.DataArray(H37)
        final_xarray["V89"] = xr.DataArray(V89)
        final_xarray["H89"] = xr.DataArray(H89)
        final_xarray["V166"] = xr.DataArray(V166)
        final_xarray["H166"] = xr.DataArray(H166)
        final_xarray["V183-3"] = xr.DataArray(V183_3)
        final_xarray["V183-7"] = xr.DataArray(V183_7)
        final_xarray["time"] = xr.DataArray(
            pd.DataFrame(time_scan)
            .astype(int)
            .apply(pd.to_datetime, format="%Y%m%d%H%M%S")
        )
    else:
        final_xarray["latitude"] = xr.DataArray(
            np.vstack([xarray_gmi["latitude"].to_masked_array(), lat])
        )
        final_xarray["longitude"] = xr.DataArray(
            np.vstack([xarray_gmi["longitude"].to_masked_array(), lon])
        )
        final_xarray["V10"] = xr.DataArray(
            np.vstack([xarray_gmi["V10"].to_masked_array(), V10])
        )
        final_xarray["H10"] = xr.DataArray(
            np.vstack([xarray_gmi["H10"].to_masked_array(), H10])
        )
        final_xarray["V19"] = xr.DataArray(
            np.vstack([xarray_gmi["V19"].to_masked_array(), V19])
        )
        final_xarray["H19"] = xr.DataArray(
            np.vstack([xarray_gmi["H19"].to_masked_array(), H19])
        )
        final_xarray["V23"] = xr.DataArray(
            np.vstack([xarray_gmi["V23"].to_masked_array(), V23])
        )
        final_xarray["V37"] = xr.DataArray(
            np.vstack([xarray_gmi["V37"].to_masked_array(), V37])
        )
        final_xarray["H37"] = xr.DataArray(
            np.vstack([xarray_gmi["H37"].to_masked_array(), H37])
        )
        final_xarray["V89"] = xr.DataArray(
            np.vstack([xarray_gmi["V89"].to_masked_array(), V89])
        )
        final_xarray["H89"] = xr.DataArray(
            np.vstack([xarray_gmi["H89"].to_masked_array(), H89])
        )
        final_xarray["V166"] = xr.DataArray(
            np.vstack([xarray_gmi["V166"].to_masked_array(), V166])
        )
        final_xarray["H166"] = xr.DataArray(
            np.vstack([xarray_gmi["H166"].to_masked_array(), H166])
        )
        final_xarray["V183-3"] = xr.DataArray(
            np.vstack([xarray_gmi["V183-3"].to_masked_array(), V183_3])
        )
        final_xarray["V183-7"] = xr.DataArray(
            np.vstack([xarray_gmi["V183-7"].to_masked_array(), V183_7])
        )
        new_time = xr.DataArray(
            pd.DataFrame(time_scan)
            .astype(int)
            .apply(pd.to_datetime, format="%Y%m%d%H%M%S")
        )
        final_xarray["time"] = xr.DataArray(
            np.vstack(
                [
                    xarray_gmi["time"].to_masked_array(),
                    new_time.to_masked_array(),
                ]
            )
        )
    return final_xarray


def call(fnames, metadata_only=False, chans=None, area_def=None, self_register=False):
    """Read GMI hdf5 data products.

    Parameters
    ----------
    fnames : list
        * List of strings, full paths to files
    metadata_only : bool, default=False
        * NOT YET IMPLEMENTED
        * Return before actually reading data if True
    chans : list of str, default=None
        * NOT YET IMPLEMENTED
        * List of desired channels (skip unneeded variables as needed).
        * Include all channels if None.
    area_def : pyresample.AreaDefinition, default=None
        * NOT YET IMPLEMENTED
        * Specify region to read
        * Read all data if None.
    self_register : str or bool, default=False
        * NOT YET IMPLEMENTED
        * register all data to the specified dataset id (as specified in the
          return dictionary keys).
        * Read multiple resolutions of data if False.

    Returns
    -------
    dict of xarray.Datasets
        * dictionary of xarray.Dataset objects with required Variables and
          Attributes.
        * Dictionary keys can be any descriptive dataset ids.

    See Also
    --------
    :ref:`xarray_standards`
        Additional information regarding required attributes and variables
        for GeoIPS-formatted xarray Datasets.
    """
    # fname='data_gmi/20200518.203639.gpm.gmi.gpm_pps.x.gmi.TB2016.x.TB2016_1b_v05a.h5'

    LOG.info("Reading files %s", fnames)

    xarray_gmi = xr.Dataset()
    source_file_names = []
    for fname in fnames:
        source_file_names += [basename(fname)]
        xarray_gmi = read_gmi_file(fname, xarray_gmi)

    # setup attributors
    from geoips.xarray_utils.time import (
        get_max_from_xarray_time,
        get_min_from_xarray_time,
    )

    xarray_gmi.attrs["source_file_names"] = sorted(source_file_names)
    xarray_gmi.attrs["start_datetime"] = get_min_from_xarray_time(xarray_gmi, "time")
    xarray_gmi.attrs["end_datetime"] = get_max_from_xarray_time(xarray_gmi, "time")
    xarray_gmi.attrs["source_name"] = "gmi"
    xarray_gmi.attrs["platform_name"] = "GPM"
    xarray_gmi.attrs["data_provider"] = "NASA"
    xarray_gmi.attrs["granule_minutes"] = 5

    # MTIFs need to be "prettier" for PMW products, so 2km resolution for final image
    xarray_gmi.attrs["sample_distance_km"] = 2
    xarray_gmi.attrs["interpolation_radius_of_influence"] = 12500

    return {"GMI": xarray_gmi, "METADATA": xarray_gmi[[]]}


def get_test_files(test_data_dir):
    """Yeilds unit files and test xarray for unit testing."""
    filepath = test_data_dir + "/test_data_gpm/data/1B*.RT-H5"
    filelist = glob(filepath)
    tmp_xr = call(filelist)
    if len(filelist) == 0:
        raise NameError("No files found")
    return tmp_xr


def get_test_parameters():
    """Yeilds data key for unit testing."""
    return [{"data_key": "GMI", "data_var": "V10"}]



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Takes in a dictionary of xarrays and converts to xarray datatree."""


import pytest
import xarray
import numpy as np
from geoips.commandline.log_setup import setup_logging
from geoips.xarray_utils import xr_to_dtree

LOG = setup_logging()


@pytest.fixture
def load_testfiles():
    """Generate test xarrays for testing."""
    # only want a small sample for testing
    xarray_dict = {
        "test_{}".format(k): xarray.Dataset(
            data_vars=dict(temp=(["x", "y"], np.empty((2, 2)))),
            coords=dict(lon=(["x", "y"], np.empty((2, 2)))),
        )
        for k in range(0, 4)
    }
    xarray_dict["METADATA"] = xarray.Dataset(
        attrs={"units": "test", "dataset": "synthetic"}
    )
    return xarray_dict


def test_xarray_to_datatree(load_testfiles):
    """Test conversion a xarray dictionary to a datatree.

    Parameters
    ----------
    load_testfiles : dict(xarrays)
        Dictionary of xarrays

    Returns
    -------
    None
    """
    metadata_copy = load_testfiles["METADATA"].copy()
    xarray_datatree = xr_to_dtree.xarray_to_datatree(load_testfiles)
    assert xarray_datatree.attrs == metadata_copy.attrs
    assert xarray_datatree.keys() == load_testfiles.keys()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Sector spec generators interface module."""

from geoips.interfaces.base import BaseModuleInterface


class SectorSpecGeneratorsInterface(BaseModuleInterface):
    """Interface for generating appropriate specifications for a sector.

    Generates a sector specification (location, resolution, etc)
    using a dynamically determined subset of information (ie, center
    lat/lon of the region determined at run time, while the resolution
    and width/height are predetermined)
    """

    name = "sector_spec_generators"
    required_args = {"area_definition": []}
    required_kwargs = {"area_definition": []}


sector_spec_generators = SectorSpecGeneratorsInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Utility for estimating the area extent, used in pyresample area definitions."""

import argparse
import numpy as np
from string import Template
import logging

LOG = logging.getLogger(__name__)

EARTH_RADIUS_METERS = 6371228.0


def generateMinMaxLatLong(lat_0, lon_0, height, width, resolution):
    """Generate minimum and maximum latitude longitude pairs.

    Min/max lat/lon based off the resolution and height/width provided.

    Parameters
    ----------
    lat_0, lon_0 : float
        Pair of latitude and longitude coordinates in degrees

    height, width, resolution : int
        Represents pixel dimensions and resolution of image in meters

    Returns
    -------
    list of floats
        min_lat, max_lat, min_lon, max_lon
    """
    phi_1 = np.deg2rad(lat_0)
    phi_2 = np.deg2rad(lon_0)
    lat_distance = height * resolution
    lon_distance = width * resolution
    lat_rad_dist = lat_distance / EARTH_RADIUS_METERS
    lon_rad_dist = lon_distance / EARTH_RADIUS_METERS
    min_lat = np.rad2deg(phi_1 - lat_rad_dist)
    max_lat = np.rad2deg(phi_1 + lat_rad_dist)
    delta_lon = np.arcsin(np.sin(lon_rad_dist) / np.cos(phi_1))
    min_lon = np.rad2deg(phi_2 - delta_lon)
    max_lon = np.rad2deg(phi_2 + delta_lon)
    return [min_lat, max_lat, min_lon, max_lon]


def haversine_distance(lat1, lon1, lat2, lon2):
    """Calculate the distance between two latitude and longitude points.

    Uses the haversine formula.

    Parameters
    ----------
    lat1, lon1, lat2, lon2 : float
        Pair of latitude and longitude coordinates in degrees

    Returns
    -------
    float
        Distance in meters between two coordinates
    """
    # Haversine formula:
    # a = sin^2( delta_phiΔφ/2) + cos(phi_1) * cos(phi_2) * sin^2(Δdelta_lambdaλ/2)
    # c = 2 * atan2( sqrt(a), sqrt(1-a) )
    # d = R * c

    # Convert lats and lons to radians
    phi_1 = np.deg2rad(lat1)
    phi_2 = np.deg2rad(lat2)
    d_phi = np.deg2rad(lat2 - lat1)
    d_lambda = np.deg2rad(lon2 - lon1)

    # Calculate haversine distance
    a = (
        np.sin(d_phi / 2.0) ** 2
        + np.cos(phi_1) * np.cos(phi_2) * np.sin(d_lambda / 2.0) ** 2
    )
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    d = EARTH_RADIUS_METERS * c
    return d


def convert_west2east(longitude):
    """Convert Longitude from degrees West to degrees East, if applicable.

    Parameters
    ----------
    longitude : float
        Longitude (degrees West or East)

    Returns
    -------
    float
        Longitude in degrees East
    """
    if longitude < 0:
        lon_east = longitude + 360
    else:
        lon_east = longitude
    return lon_east


def center_longitude(min_longitude, max_longitude):
    """Determine the center longitude based off longitude in either degW or degE.

    Parameters
    ----------
    min_longitude, max_longitude : float
        Min and Max Longitude (degrees West or East)

    Returns
    -------
    float
        Center longitude in degrees West or East
    """
    min_lonE = convert_west2east(min_longitude)
    max_lonE = convert_west2east(max_longitude)
    if min_lonE > max_lonE:
        # Prime meridian edge case (e.g. -5W - 5E)
        center_lon = (min_longitude + max_longitude) / 2.0
    else:
        center_lon = (min_lonE + max_lonE) / 2.0
    if center_lon > 180:
        center_lon -= 360
    return center_lon


def estimate_area_extent(min_lat, min_lon, max_lat, max_lon, resolution):
    """Estimate the area extent for use in the YAML area definition.

    Parameters
    ----------
    min_lat, min_lon, max_lat, max_lon : float
        Min/Max lat/lon values in degrees
    resolution : float
        Resolution in meters

    Returns
    -------
    dict
        Dictionary holding:

        * lower_left_xy - list of projection x/y coordinates of
          lower left corner of lower left pixel
        * upper_right_xy - list of projection x/y coordinates of
          upper right corner of upper right pixel
        * height - number of grid rows
        * width - number of grid columns
        * lat_0 - Center latitude in degrees
        * lon_0 - Center longitude in degrees
    """
    if min_lat == max_lat:
        raise ValueError("MIN_LAT and MAX_LAT must be different")
    if min_lon == max_lon:
        raise ValueError("MIN_LON and MAX_LON must be different")
    # Convert all longitude points to degrees east
    min_lonE = convert_west2east(min_lon)
    max_lonE = convert_west2east(max_lon)
    center_lat = (min_lat + max_lat) / 2.0
    center_lon = center_longitude(min_lon, max_lon)
    lat_distance = haversine_distance(min_lat, center_lon, max_lat, center_lon)
    # The haversine formula will find the shortest distance between two points
    # If we want the sector larger than 180 degrees in longitude, we need
    # some special handling for the distance calculation
    if (max_lonE - min_lonE) > 180:
        dist1 = haversine_distance(center_lat, min_lonE, center_lat, min_lonE + 180)
        dist2 = haversine_distance(center_lat, min_lonE + 180, center_lat, max_lonE)
        lon_distance = dist1 + dist2
    else:
        lon_distance = haversine_distance(center_lat, min_lon, center_lat, max_lon)
    extent_lat = int(lat_distance / 2.0)
    extent_lon = int(lon_distance / 2.0)
    height = int(lat_distance / resolution)
    width = int(lon_distance / resolution)
    lower_left_xy = [-1 * extent_lon, -1 * extent_lat]
    upper_right_xy = [extent_lon, extent_lat]
    return {
        "lower_left_xy": lower_left_xy,
        "upper_right_xy": upper_right_xy,
        "height": height,
        "width": width,
        "lat_0": center_lat,
        "lon_0": center_lon,
    }


def esitmate_area_from_center(lat_0, lon_0, height, width, resolution):
    """Estimate the area extent for use in the YAML area definition.

    Parameters
    ----------
    lat_0, lon_0 : float
        Center lat/lon values in degrees
    height, width : int
        Pixel dimensions
    resolution : int
        Resolution in meters

    Returns
    -------
    dict
        Dictionary holding:

        * lower_left_xy - list of projection x/y coordinates of
          lower left corner of lower left pixel
        * upper_right_xy - list of projection x/y coordinates of
          upper right corner of upper right pixel
        * height - number of grid rows
        * width - number of grid columns
        * lat_0 - Center latitude in degrees
        * lon_0 - Center longitude in degrees
    """
    # Convert all longitude points to degrees east
    center_lon = convert_west2east(lon_0)
    center_lat = lat_0
    lats_lons = generateMinMaxLatLong(lat_0, lon_0, height, width, resolution)
    max_lonE = convert_west2east(lats_lons[3])
    min_lonE = convert_west2east(lats_lons[2])
    lat_distance = haversine_distance(
        lats_lons[0], center_lon, lats_lons[1], center_lon
    )
    # distance =
    #           inverse_haversine_distance(lat_0, center_lon, height, width, resolution)
    if (max_lonE - min_lonE) > 180:
        dist1 = haversine_distance(center_lat, min_lonE, center_lat, min_lonE + 180)
        dist2 = haversine_distance(center_lat, min_lonE + 180, center_lat, max_lonE)
        lon_distance = dist1 + dist2
    else:
        lon_distance = haversine_distance(
            center_lat, lats_lons[2], center_lat, lats_lons[3]
        )
    # The haversine formula will find the shortest distance between two points
    # If we want the sector larger than 180 degrees in longitude, we need
    # some special handling for the distance calculation
    extent_lat = int(lat_distance / 2.0)
    extent_lon = int(lon_distance / 2.0)
    lower_left_xy = [-1 * extent_lon, -1 * extent_lat]
    upper_right_xy = [extent_lon, extent_lat]
    return {
        "lower_left_xy": lower_left_xy,
        "upper_right_xy": upper_right_xy,
        "height": height,
        "width": width,
        "lat_0": lat_0,
        "lon_0": lon_0,
    }


AREA_DEF_TEMPLATE = """
# Copy and paste the following into a yaml sector file.
# Replace @SECTOR_NAME@ and @SECTOR_DESCRIPTION@ with appropriate information.
# Update sector_info as needed as well

@SECTOR_NAME@:
  description: @SECTOR_DESCRIPTION@
  projection:
      a: $EARTH_RADIUS_METERS
      lat_0: $lat_0
      lon_0: $lon_0
      proj: eqc
      units: m
  resolution:
  - $resolution
  - $resolution
  shape:
      height: $height
      width: $width
  area_extent:
      lower_left_xy: $lower_left_xy
      upper_right_xy: $upper_right_xy
"""


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    subparser = parser.add_subparsers(help="sub-parser help")
    parser_c = subparser.add_parser("c", help="center help")
    parser_e = subparser.add_parser("e", help="extent help")
    parser_e.add_argument("--max_lat", type=float, required=True)
    parser_e.add_argument("--max_lon", type=float, required=True)
    parser_e.add_argument("--min_lat", type=float, required=True)
    parser_e.add_argument("--min_lon", type=float, required=True)
    parser_e.add_argument("--resolution", "-r", type=int, required=True)
    parser_c.add_argument("--lat_0", type=float, required=True)
    parser_c.add_argument("--lon_0", type=float, required=True)
    parser_c.add_argument("--height", type=int, required=True)
    parser_c.add_argument("--width", "-w", type=int, required=True)
    parser_c.add_argument("--resolution", "-r", type=int, required=True)

    ARGS = parser.parse_args()
    try:
        estimated_extent = estimate_area_extent(
            ARGS.min_lat, ARGS.min_lon, ARGS.max_lat, ARGS.max_lon, ARGS.resolution
        )
    except argparse.ArgumentError:
        estimated_extent = esitmate_area_from_center(
            ARGS.lat_0, ARGS.lon_0, ARGS.height, ARGS.width, ARGS.resolution
        )
    lat_0 = estimated_extent["lat_0"]
    lon_0 = estimated_extent["lon_0"]
    height, width = estimated_extent["height"], estimated_extent["width"]
    lower_left_xy = estimated_extent["lower_left_xy"]
    upper_right_xy = estimated_extent["upper_right_xy"]

    populated_template = Template(AREA_DEF_TEMPLATE).substitute(
        {
            "EARTH_RADIUS_METERS": EARTH_RADIUS_METERS,
            "lat_0": lat_0,
            "lon_0": lon_0,
            "resolution": ARGS.resolution,
            "height": height,
            "width": width,
            "lower_left_xy": lower_left_xy,
            "upper_right_xy": upper_right_xy,
        }
    )
    LOG.info(populated_template)



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `test unit-test` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

# from glob import glob
# from numpy.random import rand
# from os import listdir
# from os.path import basename
# import pytest
# from importlib.resources import files

# from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


# class TestGeoipsTestUnitTest(BaseCliTest):
#     """Unit Testing Class for Test Unit Test Command."""

#     @property
#     def command_combinations(self):
#         """A stochastic list of commands used by the GeoipsTestUnitTest command.

#         This includes failing cases as well.
#         """
#         if not hasattr(self, "_cmd_list"):
#             base_args = ["geoips", "test", "unit-test"]
#             self._cmd_list = [
#                 base_args + ["commandline", "-n", "test_geoips_list_packages.py"]
#             ]
#             # select a small random amount of tests to call via geoips test unit-test
#             for pkg_name in self.plugin_package_names:
#                 unit_test_dir = str(files(pkg_name) / "../tests/unit_tests")
#                 try:
#                     listdir(unit_test_dir)
#                 except FileNotFoundError:
#                     continue
#                 for dir_name in listdir(unit_test_dir):
#                     for test_name in glob(f"{unit_test_dir}/{dir_name}/test*.py"):
#                         test_name = basename(test_name)
#                         do_geoips_test = rand() < 0.05
#                         if do_geoips_test:
#                             self._cmd_list.append(
#                                 base_args + ["-p", pkg_name, dir_name, "-n", test_name] # NOQA
#                             )
#             # Add argument list to retrieve help message
#             self._cmd_list.append(base_args + ["-h"])
#             # Add argument list with non existent package and directory name
#             self._cmd_list.append(
#                 base_args + ["-p", "non_existent_package", "non_existent_dirname"]
#             )
#         return self._cmd_list

#     def check_error(self, args, error):
#         """Ensure that the 'geoips test unit-test ...' error output is correct.

#         Parameters
#         ----------
#         args: 2D list of str
#             - The arguments used to call the CLI (expected to fail)
#         error: str
#             - Multiline str representing the error output of the CLI call
#         """
#         # An error occurred using args. Assert that args is not valid and check the
#         # output of the error.
#         err_str = "To use, type\n`geoips test unit-test -p <package_name> "
#         err_str += "<directory_name> <-n> <script_name>`"
#         assert err_str in error

#     def check_output(self, args, output):
#         """Ensure that the 'geoips unit-test ...' successful output is correct.

#         Parameters
#         ----------
#         args: 2D list of str
#             - The arguments used to call the CLI
#         output: str
#             - Multiline str representing the output of the CLI call
#         """
#         # The args provided are valid, so test that the output is actually correct
#         if "-h" in args:
#             help_str = "To use, type\n`geoips test unit-test -p <package_name> "
#             help_str += "<directory_name> <-n> <script_name>`"
#             assert help_str in output
#         else:
#             # Checking that output from geoips unit-test command reports succeeds
#             assert "test session starts" in output
#             assert "pytest" in output


# test_sub_cmd = TestGeoipsTestUnitTest()


# @pytest.mark.parametrize(
#     "args",
#     test_sub_cmd.command_combinations,
#     ids=test_sub_cmd.generate_id,
# )
# def test_command_combinations(monkeypatch, args):
#     """Test all 'geoips test unit-test ...' commands.

#     This test covers every valid combination of commands for the 'geoips test unit-test' # NOQA
#     command. We also test invalid commands, to ensure that the proper help documentation # NOQA
#     is provided for those using the command incorrectly.

#     Parameters
#     ----------
#     args: 2D array of str
#         - List of arguments to call the CLI with (ie. ['geoips', 'test', 'unit-test'])
#     """
#     test_sub_cmd.test_command_combinations(monkeypatch, args)

# THIS SECTION WAS COMMENTED OUT BECAUSE WE DON"T WANT TO INVOKE PYTEST IN MAIN GEOIPS
# AND THIS COMMAND IS LARGELY FOR EASE OF USE'S SAKE. WE MAY WANT TO USE IT LATER SO
# THAT IS WHY IT HAS NOT BEEN DELETED ALTOGETHER.



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Readers interface module."""

import collections
from datetime import datetime
from os.path import basename

import numpy as np
from xarray import concat, Dataset

from geoips.errors import NoValidFilesError
from geoips.interfaces.base import BaseModuleInterface
from geoips.plugins.modules.readers.utils.hrit_reader import HritError


class ReadersInterface(BaseModuleInterface):
    """Interface for ingesting a specific data type.

    Provides specification for ingensting a specific data type, and storing in
    the GeoIPS xarray-based internal format.
    """

    name = "readers"
    required_args = {"standard": ["fnames"]}
    required_kwargs = {
        "standard": ["metadata_only", "chans", "area_def", "self_register"]
    }

    def read_data_to_xarray_dict(
        self,
        fnames,
        read_single_time_func,
        metadata_only=False,
        chans=None,
        area_def=None,
        self_register=False,
    ):
        """Read in data potentially from multiple scan times into an xarray dict.

        This function does not require that you provide multiple scan times, but allows
        for that in the case those are provided.

        Call this with information specific to your reader to generate a dictionary of
        xarray datasets created from the data provided in 'fnames'.

        Parameters
        ----------
        fnames : list
            * List of strings, full paths to files
        all_metadata : dict
            * Dictionary of metadata from all files in 'fnames'
        read_single_time_func : python function
            * Function which can be used to read a single scan time of files from a
              reader plugin.
            * Most likely named 'call_single_time'.
        metadata_only : bool, default=False
            * Return before actually reading data if True
        chans : list of str, default=None
            * List of desired channels (skip unneeded variables as needed).
            * Include all channels if None.
        area_def : pyresample.AreaDefinition, default=None
            * Specify region to read
            * Read all data if None.
        self_register : str or bool, default=False
            * register all data to the specified dataset id (as specified in the
              return dictionary keys).
            * Read multiple resolutions of data if False.

        Returns
        -------
        dict of xarray.Datasets
            * dictionary of xarray.Dataset objects with required Variables and
              Attributes.
            * Dictionary keys can be any descriptive dataset ids.

        See Also
        --------
        :ref:`xarray_standards`
            Additional information regarding required attributes and variables
            for GeoIPS-formatted xarray Datasets.
        """
        # Sort fnames. This should sort them by time and channel more often than not.
        fnames = sorted(fnames)
        # We really only need to get the start time from all of these files. I wish
        # there were an easier and more efficient route, but this should work for the
        # time being. Personally, I think it'd be a good idea to add a new argument to
        # this method that accepts a function which calculates the start datetime of
        # all files provided.
        all_file_metadata = []
        updated_fnames = []
        for x in fnames:
            try:
                all_file_metadata.append(
                    read_single_time_func([x], metadata_only=True, chans=chans)[
                        "METADATA"
                    ]
                )
                updated_fnames += [x]
            except NoValidFilesError:
                # If the current file is not valid, just skip this one.
                # print(f"{str(e)}: No files found, skipping")
                continue
            except (ValueError, HritError) as e:
                # Value error is raised for 'inconsistent' metadata, or in this case
                # No appropriate file for the channels selected
                if isinstance(e, ValueError):
                    st = None
                    et = None
                else:
                    """
                    This occurs from the seviri_hrit reader in 'get_top_level_metadata'
                    Parse out the start and end datetimes, as this file still could be
                    Relevant, but is missing 'block_2'. If the set of files all are
                    missing block_2, or the projection of block_2 is not GEOS, it will
                    cause an HritError in the for loop before, which doesn't have a try
                    except statement.

                    Error Format:
                    f"Unknown projection encountered: {projection}.\n"
                    f"start_datetime={st.isoformat()}\n"
                    f"end_datetime={et.isoformat()}"
                    """
                    emsg = str(e).split("\n")
                    # Recreate the datetime objects from the isoformat strings provided
                    st = datetime.fromisoformat(emsg[1].split("=")[1])
                    et = datetime.fromisoformat(emsg[2].split("=")[1])
                # Add st, et as datetimes for the file, nonetheless if they are None
                # or a valid datetime
                all_file_metadata.append(
                    Dataset(attrs=dict(start_datetime=st, end_datetime=et))
                )
                updated_fnames += [x]
        self.start_times = [md.attrs["start_datetime"] for md in all_file_metadata]
        self.end_times = [md.attrs["end_datetime"] for md in all_file_metadata]

        # Get unique start times and end times. We do this by initally creating a set
        # from all found start and end times. Keep in mind, some of these times could
        # be 'None' if a value error is raised from the reader call function. This
        # occurs when none of the selected channels were found in the file provided.
        # Meaning, we don't need that file..
        # Now, remove 'None' from the set (the files the don't need). This leaves a list
        # of unique file times that have been found in datasets that contain the correct
        # channels.
        self.unique_stimes = list(set(self.start_times).difference(set([None])))
        self.unique_etimes = list(set(self.end_times).difference(set([None])))
        # Set these values to this class so they can be used downstream for reading
        # data from the correct time steps
        metadata_by_scan_time = []
        for stime in self.unique_stimes:
            # Get the indices of all files which match the current start time
            same_scan_time_files = [dt == stime for dt in self.start_times]
            # Now get the metadata of all of the files which match that time
            metadata_by_scan_time.append(
                read_single_time_func(
                    list(np.array(updated_fnames)[same_scan_time_files]),
                    metadata_only=True,
                    chans=chans,
                )["METADATA"]
            )
        all_metadata = self.concatenate_metadata(metadata_by_scan_time)
        if metadata_only:
            return all_metadata

        dict_xarrays = self.call_files_and_get_top_level_metadata(
            updated_fnames,
            all_metadata,
            read_single_time_func,
            metadata_only,
            chans,
            area_def,
            self_register,
        )
        return dict_xarrays

    def concatenate_metadata(self, all_metadata):
        """Merge together metadata sourced from a list of files into one dictionary.

        Where the structure of the merged metadata is a nested dictionary of metadata.

        Ie. (xarray_obj has no data and is merely just a container for metadata):
        {"METADATA": xobj.source_file_attributes: {fname: xobj, ..., "fnamex": xobj}}

        Parameters
        ----------
        all_metadata: list of xarray.Datasets
            - The incoming metadata from X number of files

        Returns
        -------
        md: dict of xarray Datasets
            - All metadata merged into a dictionary of xarray Datasets
        """
        md = {"METADATA": Dataset()}

        for md_idx in range(len(all_metadata)):
            # Set required attributes of the top-level metadata when this loop is
            # started
            if md_idx == 0:
                md["METADATA"].attrs = all_metadata[md_idx].attrs
                md["METADATA"].attrs["source_file_names"] = []
                md["METADATA"].attrs["source_file_attributes"] = {}
                md["METADATA"].attrs["source_file_datetimes"] = []
                md["METADATA"].attrs["end_datetime"] = all_metadata[-1].end_datetime

            # Add to optional attributes of the top-level metadata for each xobj
            # provided
            md["METADATA"].attrs["source_file_names"] += [
                basename(x) for x in all_metadata[md_idx].attrs["source_file_names"]
            ]
            md["METADATA"].attrs["source_file_datetimes"].append(
                [
                    all_metadata[md_idx].start_datetime,
                    all_metadata[md_idx].end_datetime,
                ],
            )
            for x in all_metadata[md_idx].attrs["source_file_names"]:
                md["METADATA"].attrs["source_file_attributes"][basename(x)] = (
                    all_metadata[md_idx]
                )

        return md

    def call_files_and_get_top_level_metadata(
        self,
        fnames,
        all_metadata,
        call_single_file_func,
        metadata_only=False,
        chans=None,
        area_def=None,
        self_register=False,
    ):
        """
        Read in data from a list of filenames.

        Parameters
        ----------
        fnames : list
            * List of strings, full paths to files
        all_metadata : dict
            * Dictionary of metadata from all files in 'fnames'
        read_single_time_func : python function
            * Function which can be used to read a single scan time of files from a
              reader plugin.
            * Most likely named 'call_single_time'.
        metadata_only : bool, default=False
            * Return before actually reading data if True
        chans : list of str, default=None
            * List of desired channels (skip unneeded variables as needed).
            * Include all channels if None.
        area_def : pyresample.AreaDefinition, default=None
            * Specify region to read
            * Read all data if None.
        self_register : str or bool, default=False
            * register all data to the specified dataset id (as specified in the
              return dictionary keys).
            * Read multiple resolutions of data if False.

        Returns
        -------
        dict of xarray.Datasets
            * dictionary of xarray.Dataset objects with required Variables and
              Attributes.
            * Dictionary keys can be any descriptive dataset ids.

        See Also
        --------
        :ref:`xarray_standards`
            Additional information regarding required attributes and variables
            for GeoIPS-formatted xarray Datasets.
        """
        ingested_xarrays = collections.defaultdict(list)
        for time in self.unique_stimes:
            scan_time_files = [dt == time for dt in self.start_times]
            # Call the associated reader for a series of files associated with the same
            # scan time
            data_dict = call_single_file_func(
                list(np.array(fnames)[scan_time_files]),
                metadata_only=metadata_only,
                chans=chans,
                area_def=area_def,
                self_register=self_register,
            )
            for (
                dname,
                dset,
            ) in data_dict.items():
                ingested_xarrays[dname].append(dset)

        if len(self.unique_stimes) == 1:
            # No need to stack if we are only reading in one scan time
            # This is likely temporary to maintain backwards compatibility

            # This is not hit if we are provided multiple scan times.
            return data_dict

        # Now that we've ingested all scan times, stack along time dimension
        metadata = all_metadata["METADATA"]
        dict_xarrays = {}
        for dname, list_xarrays in ingested_xarrays.items():
            if dname == "METADATA":
                continue
            merged_dset = concat(list_xarrays, dim="time")
            merged_dset.attrs["start_datetime"] = min(self.unique_stimes)
            merged_dset.attrs["end_datetime"] = max(self.unique_etimes)
            merged_dset = merged_dset.assign_coords({"time": self.unique_stimes})
            dict_xarrays[dname] = merged_dset
            # Override source_file_* attributes with what's set in all_metadata.
            dict_xarrays[dname].attrs["source_file_names"] = metadata.attrs[
                "source_file_names"
            ]
            dict_xarrays[dname].attrs["source_file_attributes"] = metadata.attrs[
                "source_file_attributes"
            ]
            dict_xarrays[dname].attrs["source_file_datetimes"] = metadata.attrs[
                "source_file_datetimes"
            ]

        metadata.attrs["start_datetime"] = min(self.unique_stimes)
        metadata.attrs["end_datetime"] = max(self.unique_etimes)
        dict_xarrays["METADATA"] = metadata

        return dict_xarrays


readers = ReadersInterface()



# # # This source code is subject to the license referenced at
# # # https://github.com/NRLMMD-GEOIPS.

"""Unit test for GeoIPS CLI `tree` command.

See geoips/commandline/ancillary_info/cmd_instructions.yaml for more information.
"""

# import pytest

# from tests.unit_tests.commandline.cli_top_level_tester import BaseCliTest


# class TestGeoipsTree(BaseCliTest):
#     """Unit Testing Class for Tree Command."""

#     @property
#     def command_combinations(self):
#         """A list of every possible call signature for the GeoipsTree command.

#         This includes failing cases as well.
#         """
#         if not hasattr(self, "_cmd_list"):
#             base_args = ["geoips", "tree"]
#             self._cmd_list = [base_args]
#             for depth in range(1, 4):
#                 self._cmd_list.append(base_args + ["--max-depth", str(depth)])
#                 self._cmd_list.append(
#                     base_args + ["--max-depth", str(depth), "--color"]
#                 )
#                 self._cmd_list.append(
#                     base_args + ["--max-depth", str(depth), "--short-name"]
#                 )
#                 self._cmd_list.append(
#                     base_args + ["--max-depth", str(depth), "--short-name", "--color"]
#                 )
#             self._cmd_list.append(base_args + ["--short-name"])
#             self._cmd_list.append(base_args + ["--color"])
#             self._cmd_list.append(base_args + ["--color", "--short-name"])
#             # Add argument list to retrieve help message
#             self._cmd_list.append(base_args + ["-h"])
#             # Add argument list with invalid depth
#             self._cmd_list.append(base_args + ["--max-depth", "-1"])
#         return self._cmd_list

#     @property
#     def command_tree(self):
#         """A depth-wise mapping of all GeoIPS CLI commands."""
#         if not hasattr(self, "_command_tree"):
#             self._command_tree = {
#                 "geoips": {
#                     "config": ["install"],
#                     "get": ["family", "interface", "package", "plugin"],
#                     "list": [
#                         "interface",
#                         "interfaces",
#                         "packages",
#                         "plugins",
#                         "scripts",
#                         "test-datasets",
#                         "unit-tests",
#                     ],
#                     "run": ["single_source", "data_fusion", "config_based"],
#                     "test": ["linting", "script"],
#                     "tree": [],
#                     "validate": [],
#                 }
#             }
#         return self._command_tree

#     def check_error(self, args, error):
#         """Ensure that the 'geoips tree ...' error output is correct.

#         Parameters
#         ----------
#         args: 2D list of str
#             - The arguments used to call the CLI (expected to fail)
#         error: str
#             - Multiline str representing the error output of the CLI call
#         """
#         # An error occurred using args. Assert that args is not valid and check the
#         # output of the error.
#         err_str = (
#             "To use, type `geoips tree <--max-depth> <int> <--color> <--short-name>`."
#         )
#         assert err_str in error
#         assert args != ["geoips", "tree"]

#     def check_output(self, args, output):
#         """Ensure that the 'geoips tree ...' successful output is correct.

#         Parameters
#         ----------
#         args: 2D list of str
#             - The arguments used to call the CLI
#         output: str
#             - Multiline str representing the output of the CLI call
#         """
#         # Needed as pytest overrides 'geoips' in 'geoips tree' when called
#         # using monkeypatch.
#         output = output.replace("pytest", "geoips")
#         # The args provided are valid, so test that the output is actually correct
#         if "-h" in args or ("--max-depth" in args and "0" in args):
#             usg_str = (
#                 "To use, type `geoips tree <--max-depth> <int> <--color> "
#                 "<--short-name>`."
#             )
#             assert usg_str in output or args == ["geoips", "tree", "--max-depth", "0"]
#         else:
#             # Checking that output from geoips tree command is valid
#             if set(["--max-depth", "1"]) <= set(args):
#                 depth = 1
#             else:
#                 depth = 2

#             short = "--short-name" in args
#             assert "geoips" in output

#             if depth > 0:
#                 for cmd_name in self.command_tree["geoips"]:
#                     if not short:
#                         # full name of command; ie 'geoips list'
#                         assert f"geoips-{cmd_name}" in output
#                     else:
#                         assert cmd_name in output
#                     if depth == 2:
#                         for subcmd_name in self.command_tree["geoips"][cmd_name]:
#                             if not short:
#                                 # full name of command; ie 'geoips list packages'
#                                 assert f"geoips-{cmd_name}-{subcmd_name}" in output
#                             else:
#                                 assert subcmd_name in output


# test_sub_cmd = TestGeoipsTree()


# @pytest.mark.parametrize(
#     "args",
#     test_sub_cmd.command_combinations,
#     ids=test_sub_cmd.generate_id,
# )
# def test_command_combinations(monkeypatch, args):
#     """Test all 'geoips tree ...' commands.

#     This test covers every valid combination of commands for the 'geoips tree'
#     command. We also test invalid commands, to ensure that the proper help documentation  # NOQA
#     is provided for those using the command incorrectly.

#     Parameters
#     ----------
#     args: 2D array of str
#         - List of arguments to call the CLI with (ie. ['geoips', 'tree'])
#     """
#     test_sub_cmd.test_command_combinations(monkeypatch, args)


# NOTE: THIS FILE IS COMMENTED OUT UNTIL THE CLI PRS HAVE BEEN MERGED AND WE CAN MAKE
# ONE FINAL UPDATE THAT WILL ENCAPSULATE ALL NEW FUNCTIONALITY
